"""Test different server responses."""

import string
from random import choice
from urllib.parse import quote

import pytest

import moto.server as server


@pytest.fixture(name="bucket_name")
def table_bucket_name() -> str:
    prefix = "table-bucket"
    random_tag = "".join(choice(string.ascii_letters) for _ in range(10))
    return (prefix + random_tag).lower()


def test_s3tables_list():
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.get("/buckets")

    assert resp.status_code == 200
    assert "tableBuckets" in resp.get_json()


def test_s3tables_create_bucket(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    assert "arn" in resp.get_json()
    assert resp.get_json()["arn"].endswith(bucket_name)


def test_s3tables_get_bucket(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    arn = resp.get_json()["arn"]

    quoted_arn = quote(arn, safe="")

    resp = test_client.get(f"/buckets/{quoted_arn}")
    assert resp.status_code == 200


def test_s3tables_delete_bucket(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    arn = resp.get_json()["arn"]

    quoted_arn = quote(arn, safe="")

    resp = test_client.delete(f"/buckets/{quoted_arn}")
    assert resp.status_code == 200


def test_s3tables_create_namespace(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    arn = resp.get_json()["arn"]

    quoted_arn = quote(arn, safe="")
    resp = test_client.put(f"/namespaces/{quoted_arn}", json={"namespace": ["bar"]})

    assert resp.status_code == 200


def test_s3tables_create_table(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    arn = resp.get_json()["arn"]

    quoted_arn = quote(arn, safe="")
    resp = test_client.put(f"/namespaces/{quoted_arn}", json={"namespace": ["bar"]})

    resp = test_client.put(
        f"/tables/{quoted_arn}/bar", json={"name": "baz", "format": "ICEBERG"}
    )
    assert resp.status_code == 200


def test_s3tables_list_tables(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    arn = resp.get_json()["arn"]

    quoted_arn = quote(arn, safe="")
    resp = test_client.put(f"/namespaces/{quoted_arn}", json={"namespace": ["bar"]})

    resp = test_client.put(
        f"/tables/{quoted_arn}/bar", json={"name": "baz", "format": "ICEBERG"}
    )

    resp = test_client.get(f"/tables/{quoted_arn}")
    assert resp.get_json()["tables"]


def test_s3tables_get_table(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    arn = resp.get_json()["arn"]

    quoted_arn = quote(arn, safe="")
    resp = test_client.put(f"/namespaces/{quoted_arn}", json={"namespace": ["bar"]})

    resp = test_client.put(
        f"/tables/{quoted_arn}/bar", json={"name": "baz", "format": "ICEBERG"}
    )

    resp = test_client.get(f"/tables/{quoted_arn}/bar/baz")
    assert resp.status_code == 200


def test_s3tables_delete_table(bucket_name: str):
    backend = server.create_backend_app("s3tables")
    test_client = backend.test_client()

    resp = test_client.put("/buckets", json={"name": bucket_name})
    arn = resp.get_json()["arn"]

    quoted_arn = quote(arn, safe="")
    resp = test_client.put(f"/namespaces/{quoted_arn}", json={"namespace": ["bar"]})

    resp = test_client.put(
        f"/tables/{quoted_arn}/bar", json={"name": "baz", "format": "ICEBERG"}
    )

    resp = test_client.delete(f"/tables/{quoted_arn}/bar/baz")
    assert resp.status_code == 200



"""lexv2models base URL and path."""

from .responses import LexModelsV2Response

url_bases = [
    r"https?://lex\.(.+)\.amazonaws\.com",
    r"https?://models-v2-lex\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/bots/$": LexModelsV2Response.dispatch,
    "{0}/bots/(?P<botId>[^/]+)/$": LexModelsV2Response.dispatch,
    "{0}/bots/(?P<botId>[^/]+)/botaliases/$": LexModelsV2Response.dispatch,
    "{0}/bots/(?P<botId>[^/]+)/botaliases/(?P<botAliasId>[^/]+)/$": LexModelsV2Response.dispatch,
    "{0}/policy/(?P<resourceArn>[^/]+)/$": LexModelsV2Response.dispatch,
    "{0}/tags/(?P<resourceARN>[^/]+)$": LexModelsV2Response.dispatch,
    "{0}/tags/(?P<arn_prefix>[^/]+)/(?P<bot_id>[^/]+)$": LexModelsV2Response.dispatch,
}



"""Unit tests for quicksight-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "Email": "fakeemail@example.com",
            "UserName": "tfacctestm9hpsr970z",
            "UserRole": "READER",
            "IdentityType": "QUICKSIGHT",
        },
        {
            "Email": "authoremail@example.com",
            "UserName": "authoremail@example.com",
            "UserRole": "AUTHOR",
            "IdentityType": "IAM",
        },
    ],
)
@mock_aws
def test_register_user__quicksight(request_params):
    client = boto3.client("quicksight", region_name="us-east-2")
    resp = client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"],
        IdentityType=request_params["IdentityType"],
        UserName=request_params["UserName"],
        UserRole=request_params["UserRole"],
    )

    assert "UserInvitationUrl" in resp
    assert "User" in resp

    assert resp["User"]["Arn"] == (
        f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:user/default/{request_params['UserName']}"
    )
    assert resp["User"]["UserName"] == request_params["UserName"]
    assert resp["User"]["Email"] == request_params["Email"]
    assert resp["User"]["Role"] == request_params["UserRole"]
    assert resp["User"]["IdentityType"] == request_params["IdentityType"]
    assert resp["User"]["Active"] is False


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "Email": "fakeemail@example.com",
            "UserName": "tfacctestm9hpsr970z",
            "UserRole": "READER",
            "IdentityType": "QUICKSIGHT",
        },
        {
            "Email": "authoremail@example.com",
            "UserName": "authoremail@example.com",
            "UserRole": "AUTHOR",
            "IdentityType": "IAM",
        },
    ],
)
@mock_aws
def test_describe_user__quicksight(request_params):
    client = boto3.client("quicksight", region_name="us-east-1")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"],
        IdentityType=request_params["IdentityType"],
        UserName=request_params["UserName"],
        UserRole=request_params["UserRole"],
    )

    resp = client.describe_user(
        UserName=request_params["UserName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    assert "User" in resp

    assert resp["User"]["Arn"] == (
        f"arn:aws:quicksight:us-east-1:{ACCOUNT_ID}:user/default/{request_params['UserName']}"
    )
    assert resp["User"]["UserName"] == request_params["UserName"]
    assert resp["User"]["Email"] == request_params["Email"]
    assert resp["User"]["Role"] == request_params["UserRole"]
    assert resp["User"]["IdentityType"] == request_params["IdentityType"]
    assert resp["User"]["Active"] is False


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "Email": "fakeemail@example.com",
            "UserName": "tfacctestm9hpsr970z",
            "UserRole": "READER",
            "IdentityType": "QUICKSIGHT",
        },
        {
            "Email": "authoremail@example.com",
            "UserName": "authoremail@example.com",
            "UserRole": "AUTHOR",
            "IdentityType": "IAM",
        },
    ],
)
@mock_aws
def test_delete_user__quicksight(request_params):
    client = boto3.client("quicksight", region_name="us-east-2")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"],
        IdentityType=request_params["IdentityType"],
        UserName=request_params["UserName"],
        UserRole=request_params["UserRole"],
    )

    client.delete_user(
        UserName=request_params["UserName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    with pytest.raises(ClientError) as exc:
        client.describe_user(
            UserName=request_params["UserName"],
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "Email": "fakeemail@example.com",
            "UserName": "tfacctestm9hpsr970z",
            "UserRole": "READER",
            "IdentityType": "QUICKSIGHT",
        },
        {
            "Email": "authoremail@example.com",
            "UserName": "authoremail@example.com",
            "UserRole": "AUTHOR",
            "IdentityType": "IAM",
        },
    ],
)
@mock_aws
def test_update_user(request_params):
    client = boto3.client("quicksight", region_name="us-west-2")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"],
        IdentityType=request_params["IdentityType"],
        UserName=request_params["UserName"],
        UserRole=request_params["UserRole"],
    )

    resp = client.update_user(
        UserName=request_params["UserName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"] + "_modified",
        Role="AUTHOR",
    )
    assert resp["User"]["Email"] == request_params["Email"] + "_modified"

    resp = client.describe_user(
        UserName=request_params["UserName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    assert "User" in resp
    assert resp["User"]["Arn"] == (
        f"arn:aws:quicksight:us-west-2:{ACCOUNT_ID}:user/default/{request_params['UserName']}"
    )
    assert resp["User"]["UserName"] == request_params["UserName"]
    assert resp["User"]["Email"] == request_params["Email"] + "_modified"
    assert resp["User"]["Role"] == "AUTHOR"
    assert resp["User"]["IdentityType"] == request_params["IdentityType"]
    assert resp["User"]["Active"] is False


@mock_aws
def test_list_users__initial():
    client = boto3.client("quicksight", region_name="us-east-2")
    resp = client.list_users(AwsAccountId=ACCOUNT_ID, Namespace="default")

    assert resp["UserList"] == []
    assert resp["Status"] == 200


@mock_aws
def test_list_users():
    client = boto3.client("quicksight", region_name="us-east-2")
    for i in range(2):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"fakeemail{i}@example.com",
            IdentityType="QUICKSIGHT",
            UserName=f"fake{i}",
            UserRole="READER",
        )
    for i in range(2, 4):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"fakeemail{i}@example.com",
            IdentityType="IAM",
            UserName=f"fakeemail{i}@example.com",
            UserRole="AUTHOR",
        )

    resp = client.list_users(AwsAccountId=ACCOUNT_ID, Namespace="default")

    assert len(resp["UserList"]) == 4
    assert resp["Status"] == 200
    for user in resp["UserList"]:
        user.pop("PrincipalId")

    assert {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:user/default/fake0",
        "UserName": "fake0",
        "Email": "fakeemail0@example.com",
        "Role": "READER",
        "IdentityType": "QUICKSIGHT",
        "Active": False,
    } in resp["UserList"]

    assert {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:user/default/fakeemail3@example.com",
        "UserName": "fakeemail3@example.com",
        "Email": "fakeemail3@example.com",
        "Role": "AUTHOR",
        "IdentityType": "IAM",
        "Active": False,
    } in resp["UserList"]


@mock_aws
def test_list_users__paginated():
    client = boto3.client("quicksight", region_name="us-east-2")
    for i in range(125):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"fakeemail{i}@example.com",
            IdentityType="QUICKSIGHT",
            UserName=f"fake{i}",
            UserRole="READER",
        )

    # default pagesize is 100
    page1 = client.list_users(AwsAccountId=ACCOUNT_ID, Namespace="default")
    assert len(page1["UserList"]) == 100
    assert "NextToken" in page1

    # We can ask for a smaller pagesize
    page2 = client.list_users(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=15,
        NextToken=page1["NextToken"],
    )
    assert len(page2["UserList"]) == 15
    assert "NextToken" in page2

    # We could request all of them in one go
    all_users = client.list_users(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=1000,
    )
    length = len(all_users["UserList"])
    # We don't know exactly how much workspaces there are, because we are running multiple tests at the same time
    assert length >= 125


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "Email": "fakeemail@example.com",
            "UserName": "user.1",
            "UserRole": "READER",
            "IdentityType": "QUICKSIGHT",
            "GroupName": "group1",
        },
        {
            "Email": "authoremail@example.com",
            "UserName": "authoremail@example.com",
            "UserRole": "AUTHOR",
            "IdentityType": "IAM",
            "GroupName": "group@test",
        },
    ],
)
@mock_aws
def test_create_group_membership(request_params):
    client = boto3.client("quicksight", region_name="us-east-2")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"],
        IdentityType=request_params["IdentityType"],
        UserName=request_params["UserName"],
        UserRole=request_params["UserRole"],
    )
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName"],
    )

    resp = client.create_group_membership(
        MemberName=request_params["UserName"],
        GroupName=request_params["GroupName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    assert resp["GroupMember"] == {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/{request_params['GroupName']}/{request_params['UserName']}",
        "MemberName": request_params["UserName"],
    }
    assert resp["Status"] == 200


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "Email": "fakeemail@example.com",
            "UserName": "user.1",
            "UserRole": "READER",
            "IdentityType": "QUICKSIGHT",
            "GroupName": "group1",
        },
        {
            "Email": "authoremail@example.com",
            "UserName": "authoremail@example.com",
            "UserRole": "AUTHOR",
            "IdentityType": "IAM",
            "GroupName": "group@test",
        },
    ],
)
@mock_aws
def test_describe_group_membership(request_params):
    client = boto3.client("quicksight", region_name="us-east-2")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"],
        IdentityType=request_params["IdentityType"],
        UserName=request_params["UserName"],
        UserRole=request_params["UserRole"],
    )
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName"],
    )

    client.create_group_membership(
        MemberName=request_params["UserName"],
        GroupName=request_params["GroupName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    resp = client.describe_group_membership(
        MemberName=request_params["UserName"],
        GroupName=request_params["GroupName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    assert resp["GroupMember"] == {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/{request_params['GroupName']}/{request_params['UserName']}",
        "MemberName": request_params["UserName"],
    }
    assert resp["Status"] == 200


@mock_aws
def test_list_group_memberships():
    client = boto3.client("quicksight", region_name="us-east-2")
    for i in range(2):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"fakeemail{i}@example.com",
            IdentityType="QUICKSIGHT",
            UserName=f"user{i}",
            UserRole="READER",
        )
    for i in range(2, 4):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"fakeemail{i}@example.com",
            IdentityType="IAM",
            UserName=f"fakeemail{i}@example.com",
            UserRole="AUTHOR",
        )

    client.create_group(
        AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName="group1"
    )
    client.create_group(
        AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName="group2"
    )

    client.create_group_membership(
        MemberName="user0",
        GroupName="group1",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )
    client.create_group_membership(
        MemberName="fakeemail2@example.com",
        GroupName="group1",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )
    client.create_group_membership(
        MemberName="user1",
        GroupName="group2",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    resp = client.list_group_memberships(
        GroupName="group1", AwsAccountId=ACCOUNT_ID, Namespace="default"
    )

    assert len(resp["GroupMemberList"]) == 2
    assert resp["Status"] == 200

    assert {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/group1/user0",
        "MemberName": "user0",
    } in resp["GroupMemberList"]
    assert {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/group1/fakeemail2@example.com",
        "MemberName": "fakeemail2@example.com",
    } in resp["GroupMemberList"]

    resp = client.list_group_memberships(
        GroupName="group2", AwsAccountId=ACCOUNT_ID, Namespace="default"
    )

    assert len(resp["GroupMemberList"]) == 1
    assert resp["Status"] == 200

    assert {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/group2/user1",
        "MemberName": "user1",
    } in resp["GroupMemberList"]


@mock_aws
def test_list_group_memberships__after_deleting_user():
    client = boto3.client("quicksight", region_name="us-east-2")
    client.create_group(
        AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName="group1"
    )
    for i in range(2):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"fakeemail{i}@example.com",
            IdentityType="QUICKSIGHT",
            UserName=f"user{i}",
            UserRole="READER",
        )
        client.create_group_membership(
            MemberName=f"user{i}",
            GroupName="group1",
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )
    for i in range(2, 4):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"fakeemail{i}@example.com",
            IdentityType="IAM",
            UserName=f"fakeemail{i}@example.com",
            UserRole="AUTHOR",
        )
        client.create_group_membership(
            MemberName=f"fakeemail{i}@example.com",
            GroupName="group1",
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )

    resp = client.list_group_memberships(
        GroupName="group1", AwsAccountId=ACCOUNT_ID, Namespace="default"
    )
    assert len(resp["GroupMemberList"]) == 4

    client.delete_user(UserName="user1", AwsAccountId=ACCOUNT_ID, Namespace="default")
    client.delete_user(
        UserName="fakeemail2@example.com",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    resp = client.list_group_memberships(
        GroupName="group1", AwsAccountId=ACCOUNT_ID, Namespace="default"
    )
    assert len(resp["GroupMemberList"]) == 2


@mock_aws
def test_list_group_memberships__check_exceptions():
    client = boto3.client("quicksight", region_name="us-east-2")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email="fakeemail@example.com",
        IdentityType="QUICKSIGHT",
        UserName="user",
        UserRole="READER",
    )

    client.create_group(AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName="group")

    client.create_group_membership(
        MemberName="user",
        GroupName="group",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    # Test for user is member of group
    resp = client.describe_group_membership(
        MemberName="user",
        GroupName="group",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )
    assert resp["Status"] == 200
    assert resp["GroupMember"] == {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/group/user",
        "MemberName": "user",
    }

    # Test for user is not member of group
    with pytest.raises(ClientError) as exc:
        resp = client.describe_group_membership(
            MemberName="fake_user",
            GroupName="group",
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"

    # Test for non existing group
    with pytest.raises(ClientError) as exc:
        resp = client.describe_group_membership(
            MemberName="user",
            GroupName="fake_group",
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_group_memberships__paginated():
    client = boto3.client("quicksight", region_name="us-east-2")
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName="group",
    )
    for i in range(125):
        client.register_user(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Email=f"authoremail{i}@example.com",
            IdentityType="QUICKSIGHT",
            UserName=f"user.{i}",
            UserRole="AUTHOR",
        )
        client.create_group_membership(
            MemberName=f"user.{i}",
            GroupName="group",
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )

    # default pagesize is 100
    page1 = client.list_group_memberships(
        GroupName="group", AwsAccountId=ACCOUNT_ID, Namespace="default"
    )
    assert len(page1["GroupMemberList"]) == 100
    assert "NextToken" in page1

    # We can ask for a smaller pagesize
    page2 = client.list_group_memberships(
        GroupName="group",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=15,
        NextToken=page1["NextToken"],
    )
    assert len(page2["GroupMemberList"]) == 15
    assert "NextToken" in page2

    # We could request all of them in one go
    all_users = client.list_group_memberships(
        GroupName="group",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=1000,
    )
    length = len(all_users["GroupMemberList"])
    # We don't know exactly how much workspaces there are, because we are running multiple tests at the same time
    assert length >= 125


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "Email": "fakeemail@example.com",
            "UserName": "user.1",
            "UserRole": "READER",
            "IdentityType": "QUICKSIGHT",
            "GroupName1": "group1",
            "GroupName2": "group2",
            "GroupName3": "group3",
        },
        {
            "Email": "authoremail@example.com",
            "UserName": "authoremail@example.com",
            "UserRole": "AUTHOR",
            "IdentityType": "IAM",
            "GroupName1": "group1@test",
            "GroupName2": "group2@test",
            "GroupName3": "group3@test",
        },
    ],
)
@mock_aws
def test_list_user_groups(request_params):
    client = boto3.client("quicksight", region_name="us-east-2")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email=request_params["Email"],
        IdentityType=request_params["IdentityType"],
        UserName=request_params["UserName"],
        UserRole=request_params["UserRole"],
    )
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName1"],
    )
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName2"],
    )
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName3"],
    )
    client.create_group_membership(
        MemberName=request_params["UserName"],
        GroupName=request_params["GroupName1"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )
    client.create_group_membership(
        MemberName=request_params["UserName"],
        GroupName=request_params["GroupName2"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    resp = client.list_user_groups(
        UserName=request_params["UserName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    assert resp["GroupList"] == [
        {
            "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/{request_params['GroupName1']}",
            "GroupName": request_params["GroupName1"],
            "PrincipalId": ACCOUNT_ID,
        },
        {
            "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/{request_params['GroupName2']}",
            "GroupName": request_params["GroupName2"],
            "PrincipalId": ACCOUNT_ID,
        },
    ]
    assert resp["Status"] == 200


@mock_aws
def test_list_user_groups__paginate():
    client = boto3.client("quicksight", region_name="us-east-2")
    client.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email="fakeemail@example.com",
        IdentityType="QUICKSIGHT",
        UserName="fake_user",
        UserRole="READER",
    )
    for i in range(125):
        client.create_group(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            GroupName=f"group{i}",
        )

        client.create_group_membership(
            MemberName="fake_user",
            GroupName=f"group{i}",
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )

    # default pagesize is 100
    page1 = client.list_user_groups(
        UserName="fake_user",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )
    assert len(page1["GroupList"]) == 100
    assert "NextToken" in page1

    # We can ask for a smaller pagesize
    page2 = client.list_user_groups(
        UserName="fake_user",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=15,
        NextToken=page1["NextToken"],
    )
    assert len(page2["GroupList"]) == 15
    assert "NextToken" in page2

    # We could request all of them in one go
    all_users = client.list_user_groups(
        UserName="fake_user",
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=1000,
    )
    length = len(all_users["GroupList"])
    # We don't know exactly how much workspaces there are, because we are running multiple tests at the same time
    assert length >= 125


@mock_aws
def test_list_users__diff_account_region():
    ACCOUNT_ID_2 = "998877665544"
    client_us = boto3.client("quicksight", region_name="us-east-2")
    client_eu = boto3.client("quicksight", region_name="eu-west-1")
    client_us.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email="fake_us_1@example.com",
        IdentityType="QUICKSIGHT",
        UserName="fake_us_1",
        UserRole="READER",
    )
    resp = client_us.register_user(
        AwsAccountId=ACCOUNT_ID_2,
        Namespace="default",
        Email="fake_us_2@example.com",
        IdentityType="QUICKSIGHT",
        UserName="fake_us_2",
        UserRole="AUTHOR",
    )
    client_eu.register_user(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Email="fake_eu_1@example.com",
        IdentityType="IAM",
        UserName="fake_eu_1",
        UserRole="AUTHOR",
    )

    # Return Account 1, Region US
    resp = client_us.list_users(AwsAccountId=ACCOUNT_ID, Namespace="default")

    assert len(resp["UserList"]) == 1
    assert resp["Status"] == 200

    resp["UserList"][0].pop("PrincipalId")

    assert resp["UserList"][0] == {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:user/default/fake_us_1",
        "UserName": "fake_us_1",
        "Email": "fake_us_1@example.com",
        "Role": "READER",
        "IdentityType": "QUICKSIGHT",
        "Active": False,
    }

    # Return Account 2, Region US
    resp = client_us.list_users(AwsAccountId=ACCOUNT_ID_2, Namespace="default")

    assert len(resp["UserList"]) == 1
    assert resp["Status"] == 200

    resp["UserList"][0].pop("PrincipalId")

    assert resp["UserList"][0] == {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID_2}:user/default/fake_us_2",
        "UserName": "fake_us_2",
        "Email": "fake_us_2@example.com",
        "Role": "AUTHOR",
        "IdentityType": "QUICKSIGHT",
        "Active": False,
    }

    # Return Account 1, Region EU
    resp = client_eu.list_users(AwsAccountId=ACCOUNT_ID, Namespace="default")

    assert len(resp["UserList"]) == 1
    assert resp["Status"] == 200

    resp["UserList"][0].pop("PrincipalId")

    assert resp["UserList"][0] == {
        "Arn": f"arn:aws:quicksight:eu-west-1:{ACCOUNT_ID}:user/default/fake_eu_1",
        "UserName": "fake_eu_1",
        "Email": "fake_eu_1@example.com",
        "Role": "AUTHOR",
        "IdentityType": "IAM",
        "Active": False,
    }



from __future__ import annotations

import logging
from typing import Optional

from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.inline_iteration_component import (
    InlineIterationComponent,
    InlineIterationComponentEvalInput,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.inline_iterator_worker import (
    InlineIteratorWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.iterator_decl import (
    IteratorDecl,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment

LOG = logging.getLogger(__name__)


class InlineIteratorEvalInput(InlineIterationComponentEvalInput):
    pass


class InlineIterator(InlineIterationComponent):
    _eval_input: Optional[InlineIteratorEvalInput]

    def _create_worker(self, env: Environment) -> InlineIteratorWorker:
        return InlineIteratorWorker(
            work_name=self._eval_input.state_name,
            job_pool=self._job_pool,
            env=env,
            parameters=self._eval_input.parameters,
            item_selector=self._eval_input.item_selector,
        )

    @classmethod
    def from_declaration(cls, iterator_decl: IteratorDecl):
        return cls(
            query_language=iterator_decl.query_language,
            start_at=iterator_decl.start_at,
            states=iterator_decl.states,
            comment=iterator_decl.comment,
            processor_config=iterator_decl.processor_config,
        )



from typing import Dict, Optional, Set, Tuple

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service import (
    StateTaskService,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for

_ERROR_NAME_AWS: str = "DynamoDB.AmazonDynamoDBException"

_SUPPORTED_API_PARAM_BINDINGS: Dict[str, Set[str]] = {
    "getitem": {
        "Key",
        "TableName",
        "AttributesToGet",
        "ConsistentRead",
        "ExpressionAttributeNames",
        "ProjectionExpression",
        "ReturnConsumedCapacity",
    },
    "putitem": {
        "Item",
        "TableName",
        "ConditionalOperator",
        "ConditionExpression",
        "Expected",
        "ExpressionAttributeNames",
        "ExpressionAttributeValues",
        "ReturnConsumedCapacity",
        "ReturnItemCollectionMetrics",
        "ReturnValues",
    },
    "deleteitem": {
        "Key",
        "TableName",
        "ConditionalOperator",
        "ConditionExpression",
        "Expected",
        "ExpressionAttributeNames",
        "ExpressionAttributeValues",
        "ReturnConsumedCapacity",
        "ReturnItemCollectionMetrics",
        "ReturnValues",
    },
    "updateitem": {
        "Key",
        "TableName",
        "AttributeUpdates",
        "ConditionalOperator",
        "ConditionExpression",
        "Expected",
        "ExpressionAttributeNames",
        "ExpressionAttributeValues",
        "ReturnConsumedCapacity",
        "ReturnItemCollectionMetrics",
        "ReturnValues",
        "UpdateExpression",
    },
}


class StateTaskServiceDynamoDB(StateTaskService):
    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return _SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    @staticmethod
    def _error_cause_from_client_error(client_error: ClientError) -> Tuple[str, str]:
        error_code: str = client_error.response["Error"]["Code"]
        error_msg: str = client_error.response["Error"]["Message"]
        response_details = "; ".join(
            [
                "Service: AmazonDynamoDBv2",
                f"Status Code: {client_error.response['ResponseMetadata']['HTTPStatusCode']}",
                f"Error Code: {error_code}",
                f"Request ID: {client_error.response['ResponseMetadata']['RequestId']}",
                "Proxy: null",
            ]
        )
        error = f"DynamoDB.{error_code}"
        cause = f"{error_msg} ({response_details})"
        return error, cause

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            error, cause = self._error_cause_from_client_error(ex)
            error_name = CustomErrorName(error)
            return FailureEvent(
                env=env,
                error_name=error_name,
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=error,
                        cause=cause,
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )
        else:
            return FailureEvent(
                env=env,
                error_name=CustomErrorName(_ERROR_NAME_AWS),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=_ERROR_NAME_AWS,
                        cause=str(ex),  # TODO: update to report expected cause.
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ):
        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        dynamodb_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
            credentials=task_credentials,
        )
        response = getattr(dynamodb_client, api_action)(**normalised_parameters)
        response.pop("ResponseMetadata", None)
        env.stack.append(response)



import json
import os
from functools import wraps
from time import sleep
from uuid import uuid4

import boto3
from botocore.exceptions import ClientError

from moto import mock_aws


def lambda_aws_verified(func):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.

    This decorator will:
      - Create an IAM-role that can be used by AWSLambda functions table
      - Run the test
      - Delete the role
    """

    @wraps(func)
    def pagination_wrapper(**kwargs):
        role_name = "moto_test_role_" + str(uuid4())[0:6]

        allow_aws_request = (
            os.environ.get("MOTO_TEST_ALLOW_AWS_REQUEST", "false").lower() == "true"
        )

        if allow_aws_request:
            return create_role_and_test(kwargs, role_name)
        else:
            with mock_aws():
                return create_role_and_test(kwargs, role_name)

    def create_role_and_test(kwargs, role_name):
        from .test_lambda import _lambda_region

        policy_doc = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "LambdaAssumeRole",
                    "Effect": "Allow",
                    "Principal": {"Service": "lambda.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        iam = boto3.client("iam", region_name=_lambda_region)
        iam_role_arn = iam.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument=json.dumps(policy_doc),
            Path="/",
        )["Role"]["Arn"]

        try:
            # Role is not immediately available
            check_iam_role_is_ready_to_use(iam_role_arn, role_name, _lambda_region)

            kwargs["iam_role_arn"] = iam_role_arn
            resp = func(**kwargs)
        finally:
            for policy in iam.list_attached_role_policies(RoleName=role_name)[
                "AttachedPolicies"
            ]:
                iam.detach_role_policy(
                    RoleName=role_name,
                    PolicyArn=policy["PolicyArn"],
                )
            iam.delete_role(RoleName=role_name)

        return resp

    def check_iam_role_is_ready_to_use(role_arn: str, role_name: str, region: str):
        from .utilities import get_test_zip_file1

        iam = boto3.client("iam", region_name=region)

        # The waiter is normally used to wait until a resource is ready
        wait_for_role = iam.get_waiter("role_exists")
        wait_for_role.wait(RoleName=role_name)

        # HOWEVER:
        # Just because the role exists, doesn't mean it can be used by Lambda
        # The only reliable way to check if our role is ready:
        #   - try to create a function and see if it succeeds
        #
        # The alternive is to wait between 5 and 10 seconds, which is not a great solution either
        _lambda = boto3.client("lambda", region)
        for _ in range(15):
            try:
                fn_name = f"fn_for_{role_name}"
                _lambda.create_function(
                    FunctionName=fn_name,
                    Runtime="python3.11",
                    Role=role_arn,
                    Handler="lambda_function.lambda_handler",
                    Code={"ZipFile": get_test_zip_file1()},
                )
                # If it succeeded, our role is ready
                _lambda.delete_function(FunctionName=fn_name)

                return
            except ClientError:
                sleep(1)
        raise Exception(
            f"Couldn't create test Lambda FN using IAM role {role_name}, probably because it wasn't ready in time"
        )

    return pagination_wrapper



"""Pagination control model for DirectoryService."""

PAGINATION_MODEL = {
    "describe_directories": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "directory_id",
    },
    "list_tags_for_resource": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 50,
        "unique_attribute": "Key",
    },
    "describe_ldaps_settings": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
    },
    "describe_trusts": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": "trust_id",
    },
    "describe_settings": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": "Name",
    },
}

# List of directory security settings
# https://docs.aws.amazon.com/directoryservice/latest/admin-guide/ms_ad_directory_settings.html#list-ds-settings
SETTINGS_ENTRIES_MODEL = [
    {
        "Type": "Protocol",
        "Name": "PCT_1_0",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Protocol",
        "Name": "SSL_2_0",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Protocol",
        "Name": "SSL_3_0",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Protocol",
        "Name": "TLS_1_0",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Protocol",
        "Name": "TLS_1_1",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "AES_128_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "DES_56_56",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "RC2_40_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "RC2_56_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "RC2_128_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "RC4_40_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "RC4_56_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "RC4_64_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "RC4_128_128",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Cipher",
        "Name": "3DES_168_168",
        "AllowedValues": '["Enable","Disable"]',
        "AppliedValue": "Enable",
        "RequestStatus": "Default",
        "DataType": "Boolean",
    },
    {
        "Type": "Certificate-Based Authentication",
        "Name": "CERTIFICATE_BACKDATING_COMPENSATION",
        "AllowedValues": '{"MinimumValue": 1, "MaximumValue": 1577000000}',
        "AppliedValue": "600",
        "RequestStatus": "Default",
        "DataType": "DurationInSeconds",
    },
    {
        "Type": "Certificate-Based Authentication",
        "Name": "CERTIFICATE_STRONG_ENFORCEMENT",
        "AllowedValues": '["Compatibility","Full_Enforcement"]',
        "AppliedValue": "Compatibility",
        "RequestStatus": "Default",
        "DataType": "Enum",
    },
]



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

from .. import dynamodb_aws_verified


@mock_aws
def test_multiple_transactions_on_same_item():
    schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **schema
    )
    # Insert an item
    dynamodb.put_item(TableName="test-table", Item={"id": {"S": "foo"}})

    def update_email_transact(email):
        return {
            "Update": {
                "Key": {"id": {"S": "foo"}},
                "TableName": "test-table",
                "UpdateExpression": "SET #e = :v",
                "ExpressionAttributeNames": {"#e": "email_address"},
                "ExpressionAttributeValues": {":v": {"S": email}},
            }
        }

    with pytest.raises(ClientError) as exc:
        dynamodb.transact_write_items(
            TransactItems=[
                update_email_transact("test1@moto.com"),
                update_email_transact("test2@moto.com"),
            ]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "Transaction request cannot include multiple operations on one item"
    )


@pytest.mark.aws_verified
@dynamodb_aws_verified(add_range=True)
def test_transact_write_items__put_and_delete_on_same_item(table_name=None):
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")

    with pytest.raises(ClientError) as exc:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Put": {
                        "TableName": table_name,
                        "Item": {
                            "pk": {"S": "test-pk"},
                            "sk": {"S": "test-sk"},
                            "field": {"S": "test-field"},
                        },
                    }
                },
                {
                    "Delete": {
                        "TableName": table_name,
                        "Key": {
                            "pk": {"S": "test-pk"},
                            "sk": {"S": "test-sk"},
                        },
                    }
                },
            ]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "Transaction request cannot include multiple operations on one item"
    )


@pytest.mark.aws_verified
@dynamodb_aws_verified(add_range=True)
def test_transact_write_items__update_with_multiple_set_clauses(table_name=None):
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")

    with pytest.raises(ClientError) as exc:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Update": {
                        "TableName": table_name,
                        "Key": {"pk": {"S": "s"}},
                        "UpdateExpression": "SET expire_at_utc = :expire_at_utc SET is_final = :true ADD version :one",
                        "ExpressionAttributeValues": {
                            ":true": {"BOOL": True},
                            ":one": {"N": "1"},
                            ":expire_at_utc": {"S": "asdf"},
                        },
                    }
                }
            ]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == 'Invalid UpdateExpression: The "SET" section can only be used once in an update expression;'
    )

    # Note that we can do this by simply separating the statements with a comma
    dynamodb.transact_write_items(
        TransactItems=[
            {
                "Update": {
                    "TableName": table_name,
                    "Key": {"pk": {"S": "s"}, "sk": {"S": "t"}},
                    "UpdateExpression": "SET a1 = :v1, a2 = :v2",
                    "ExpressionAttributeValues": {
                        ":v1": {"S": "v1"},
                        ":v2": {"S": "v1"},
                    },
                }
            }
        ]
    )

    inventory = dynamodb.scan(TableName=table_name)["Items"]
    assert inventory == [
        {"a1": {"S": "v1"}, "sk": {"S": "t"}, "a2": {"S": "v1"}, "pk": {"S": "s"}}
    ]


@mock_aws
def test_transact_write_items__too_many_transactions():
    schema = {
        "KeySchema": [{"AttributeName": "pk", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "pk", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **schema
    )

    def update_email_transact(email):
        return {
            "Put": {
                "TableName": "test-table",
                "Item": {"pk": {"S": ":v"}},
                "ExpressionAttributeValues": {":v": {"S": email}},
            }
        }

    update_email_transact("test1@moto.com")
    with pytest.raises(ClientError) as exc:
        dynamodb.transact_write_items(
            TransactItems=[
                update_email_transact(f"test{idx}@moto.com") for idx in range(101)
            ]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "1 validation error detected at 'transactItems' failed to satisfy constraint: Member must have length less than or equal to 100."
    )


@mock_aws
def test_transact_write_items_multiple_operations_fail():
    # Setup
    schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    table_name = "test-table"
    dynamodb.create_table(TableName=table_name, BillingMode="PAY_PER_REQUEST", **schema)

    # Execute
    with pytest.raises(ClientError) as exc:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Put": {
                        "Item": {"id": {"S": "test"}},
                        "TableName": table_name,
                    },
                    "Delete": {
                        "Key": {"id": {"S": "test"}},
                        "TableName": table_name,
                    },
                }
            ]
        )
    # Verify
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "TransactItems can only contain one of Check, Put, Update or Delete"
    )


@mock_aws
def test_transact_write_items_with_empty_gsi_key():
    client = boto3.client("dynamodb", "us-east-2")

    client.create_table(
        TableName="test_table",
        KeySchema=[{"AttributeName": "unique_code", "KeyType": "HASH"}],
        AttributeDefinitions=[
            {"AttributeName": "unique_code", "AttributeType": "S"},
            {"AttributeName": "unique_id", "AttributeType": "S"},
        ],
        GlobalSecondaryIndexes=[
            {
                "IndexName": "gsi_index",
                "KeySchema": [{"AttributeName": "unique_id", "KeyType": "HASH"}],
                "Projection": {"ProjectionType": "ALL"},
            }
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
    )

    transact_items = [
        {
            "Put": {
                "Item": {"unique_code": {"S": "some code"}, "unique_id": {"S": ""}},
                "TableName": "test_table",
            }
        }
    ]

    with pytest.raises(ClientError) as exc:
        client.transact_write_items(TransactItems=transact_items)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "One or more parameter values are not valid. A value specified for a secondary index key is not supported. The AttributeValue for a key attribute cannot contain an empty string value. IndexName: gsi_index, IndexKey: unique_id"
    )


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_transaction_with_empty_key(table_name=None):
    dynamodb = boto3.resource("dynamodb", region_name="us-east-1")
    client = boto3.client("dynamodb", region_name="us-east-1")
    table = dynamodb.Table(table_name)
    table.put_item(Item={"pk": "mark", "lock": {"acquired_at": 123}})

    with pytest.raises(ClientError) as exc:
        client.transact_write_items(
            TransactItems=[
                {
                    "Update": {
                        "TableName": table_name,
                        "Key": {"pk": {"S": ""}},
                        "UpdateExpression": "SET sth = :v",
                        "ExpressionAttributeValues": {":v": {"N": "0"}},
                    }
                }
            ]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "One or more parameter values are not valid. The AttributeValue for a key attribute cannot contain an empty string value. Key: pk"
    )


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_transact_write_items__empty_expr_attr_values(table_name=None):
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")

    with pytest.raises(ClientError) as exc:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Put": {
                        "Item": {"pk": {"S": "tenant#0000001-tenant#0000001"}},
                        "ConditionExpression": "attribute_not_exists(#n0)",
                        "TableName": table_name,
                        "ExpressionAttributeNames": {"#n0": "pk"},
                        "ExpressionAttributeValues": {},
                    }
                },
            ]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert err["Message"] == "ExpressionAttributeValues must not be empty"

    with pytest.raises(ClientError) as exc:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Update": {
                        "Key": {"pk": {"S": "globals"}},
                        "UpdateExpression": "SET #0 = {'S': 'asdf'}",
                        "ExpressionAttributeNames": {"#0": "tenant_count"},
                        "ExpressionAttributeValues": {},
                        "TableName": table_name,
                    }
                },
            ]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert err["Message"] == "ExpressionAttributeValues must not be empty"



from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.distributed_item_processor import (
    DistributedItemProcessor,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.inline_item_processor import (
    InlineItemProcessor,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.item_processor_decl import (
    ItemProcessorDecl,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_component import (
    IterationComponent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.mode import (
    Mode,
)


def from_item_processor_decl(
    item_processor_decl: ItemProcessorDecl,
) -> IterationComponent:
    if item_processor_decl.processor_config.mode == Mode.Inline:
        return InlineItemProcessor(
            query_language=item_processor_decl.query_language,
            start_at=item_processor_decl.start_at,
            states=item_processor_decl.states,
            comment=item_processor_decl.comment,
            processor_config=item_processor_decl.processor_config,
        )
    elif item_processor_decl.processor_config.mode == Mode.Distributed:
        return DistributedItemProcessor(
            query_language=item_processor_decl.query_language,
            start_at=item_processor_decl.start_at,
            states=item_processor_decl.states,
            comment=item_processor_decl.comment,
            processor_config=item_processor_decl.processor_config,
        )
    else:
        raise ValueError(
            f"Unknown Map state processing mode: '{item_processor_decl.processor_config.mode}'."
        )



"""cloudhsmv2 base URL and path."""

from .responses import CloudHSMV2Response

url_bases = [
    r"https?://cloudhsm\.(.+)\.amazonaws\.com",
    r"https?://cloudhsmv2\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": CloudHSMV2Response.dispatch,
}



from typing import Final

from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_terminal import (
    JSONataTemplateValueTerminalExpression,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class Seconds(WaitFunction):
    # Seconds
    # A time, in seconds, to state_wait before beginning the state specified in the Next
    # field. You must specify time as a positive, integer value.

    def __init__(self, seconds: int):
        self.seconds: int = seconds

    def _get_wait_seconds(self, env: Environment) -> int:
        return self.seconds


class SecondsJSONata(WaitFunction):
    jsonata_template_value_terminal_expression: Final[
        JSONataTemplateValueTerminalExpression
    ]

    def __init__(
        self,
        jsonata_template_value_terminal_expression: JSONataTemplateValueTerminalExpression,
    ):
        super().__init__()
        self.jsonata_template_value_terminal_expression = (
            jsonata_template_value_terminal_expression
        )

    def _get_wait_seconds(self, env: Environment) -> int:
        # TODO: add snapshot tests to verify AWS's behaviour about non integer values.
        self.jsonata_template_value_terminal_expression.eval(env=env)
        max_items: int = int(env.stack.pop())
        return max_items



from __future__ import annotations

from moto.stepfunctions.parser.api import HistoryEventType
from moto.stepfunctions.parser.asl.component.state.state import CommonStateField
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class StateWait(CommonStateField):
    wait_function: WaitFunction

    def __init__(self):
        super().__init__(
            state_entered_event_type=HistoryEventType.WaitStateEntered,
            state_exited_event_type=HistoryEventType.WaitStateExited,
        )

    def from_state_props(self, state_props: StateProps) -> None:
        super(StateWait, self).from_state_props(state_props)
        self.wait_function = state_props.get(
            typ=WaitFunction,
            raise_on_missing=ValueError(
                f"Undefined WaitFunction for StateWait: '{self}'."
            ),
        )

    def _eval_state(self, env: Environment) -> None:
        self.wait_function.eval(env)
        if self.assign_decl:
            self.assign_decl.eval(env=env)



import logging
import threading
from typing import Final, Optional

from moto.stepfunctions.parser.api import (
    ExecutionAbortedEventDetails,
    ExecutionFailedEventDetails,
    ExecutionSucceededEventDetails,
    ExecutionTimedOutEventDetails,
    HistoryEventExecutionDataDetails,
    HistoryEventType,
)
from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.common.query_language import QueryLanguage
from moto.stepfunctions.parser.asl.component.common.timeouts.timeout import (
    TimeoutSeconds,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.program.states import States
from moto.stepfunctions.parser.asl.component.program.version import Version
from moto.stepfunctions.parser.asl.component.state.state import CommonStateField
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.eval.program_state import (
    ProgramEnded,
    ProgramError,
    ProgramState,
    ProgramStopped,
    ProgramTimedOut,
)
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.utils import TMP_THREADS
from moto.utilities.collections import select_from_typed_dict

LOG = logging.getLogger(__name__)


class Program(EvalComponent):
    query_language: QueryLanguage
    start_at: Final[StartAt]
    states: Final[States]
    timeout_seconds: Final[Optional[TimeoutSeconds]]
    comment: Final[Optional[Comment]]
    version: Final[Optional[Version]]

    def __init__(
        self,
        query_language: QueryLanguage,
        start_at: StartAt,
        states: States,
        timeout_seconds: Optional[TimeoutSeconds],
        comment: Optional[Comment] = None,
        version: Optional[Version] = None,
    ):
        self.query_language = query_language
        self.start_at = start_at
        self.states = states
        self.timeout_seconds = timeout_seconds
        self.comment = comment
        self.version = version

    def _get_state(self, state_name: str) -> CommonStateField:
        state: Optional[CommonStateField] = self.states.states.get(state_name, None)
        if state is None:
            raise ValueError(f"No such state {state}.")
        return state

    def eval(self, env: Environment) -> None:
        timeout = self.timeout_seconds.timeout_seconds if self.timeout_seconds else None
        env.next_state_name = self.start_at.start_at_name
        worker_thread = threading.Thread(target=super().eval, args=(env,), daemon=True)
        TMP_THREADS.append(worker_thread)
        worker_thread.start()
        worker_thread.join(timeout=timeout)
        is_timeout = worker_thread.is_alive()
        if is_timeout:
            env.set_timed_out()

    def _eval_body(self, env: Environment) -> None:
        try:
            while env.is_running():
                # Store the heap values at this depth for garbage collection.
                heap_values = set(env.heap.keys())

                next_state: CommonStateField = self._get_state(env.next_state_name)
                next_state.eval(env)

                # Garbage collect hanging values added by this last state.
                env.stack.clear()
                clear_heap_values = set(env.heap.keys()) - heap_values
                for heap_value in clear_heap_values:
                    env.heap.pop(heap_value, None)

        except FailureEventException as ex:
            env.set_error(error=ex.get_execution_failed_event_details())
        except Exception as ex:
            cause = f"{type(ex).__name__}({str(ex)})"
            LOG.error("Stepfunctions computation ended with exception '%s'.", cause)
            env.set_error(
                ExecutionFailedEventDetails(
                    error=StatesErrorName(
                        typ=StatesErrorNameType.StatesRuntime
                    ).error_name,
                    cause=cause,
                )
            )

        # If the program is evaluating within a frames then these are not allowed to produce program termination states.
        if env.is_frame():
            return

        program_state: ProgramState = env.program_state()
        if isinstance(program_state, ProgramError):
            exec_failed_event_details = select_from_typed_dict(
                typed_dict=ExecutionFailedEventDetails,
                obj=program_state.error or dict(),
            )
            env.event_manager.add_event(
                context=env.event_history_context,
                event_type=HistoryEventType.ExecutionFailed,
                event_details=EventDetails(
                    executionFailedEventDetails=exec_failed_event_details
                ),
            )
        elif isinstance(program_state, ProgramStopped):
            env.event_history_context.source_event_id = 0
            env.event_manager.add_event(
                context=env.event_history_context,
                event_type=HistoryEventType.ExecutionAborted,
                event_details=EventDetails(
                    executionAbortedEventDetails=ExecutionAbortedEventDetails(
                        error=program_state.error, cause=program_state.cause
                    )
                ),
            )
        elif isinstance(program_state, ProgramTimedOut):
            env.event_manager.add_event(
                context=env.event_history_context,
                event_type=HistoryEventType.ExecutionTimedOut,
                event_details=EventDetails(
                    executionTimedOutEventDetails=ExecutionTimedOutEventDetails()
                ),
            )
        elif isinstance(program_state, ProgramEnded):
            env.event_manager.add_event(
                context=env.event_history_context,
                event_type=HistoryEventType.ExecutionSucceeded,
                event_details=EventDetails(
                    executionSucceededEventDetails=ExecutionSucceededEventDetails(
                        output=to_json_str(
                            env.states.get_input(), separators=(",", ":")
                        ),
                        outputDetails=HistoryEventExecutionDataDetails(
                            truncated=False  # Always False for api calls.
                        ),
                    )
                ),
            )



import json
from typing import Any, Dict, List, Optional

from moto.core.exceptions import JsonRESTError
from moto.dynamodb.limits import HASH_KEY_MAX_LENGTH, RANGE_KEY_MAX_LENGTH

ERROR_TYPE_PREFIX = "com.amazonaws.dynamodb.v20120810#"


class DynamodbException(JsonRESTError):
    pass


class MockValidationException(DynamodbException):
    error_type = ERROR_TYPE_PREFIX + "ValidationException"

    def __init__(self, message: str):
        super().__init__(MockValidationException.error_type, message=message)
        self.exception_msg = message


class KeyIsEmptyStringException(MockValidationException):
    def __init__(self, empty_key: str):
        super().__init__(
            message=f"One or more parameter values are not valid. The AttributeValue for a key attribute cannot contain an empty string value. Key: {empty_key}"
        )


class InvalidIndexNameError(MockValidationException):
    pass


class InvalidUpdateExpressionInvalidDocumentPath(MockValidationException):
    invalid_update_expression_msg = (
        "The document path provided in the update expression is invalid for update"
    )

    def __init__(self) -> None:
        super().__init__(self.invalid_update_expression_msg)


class InvalidUpdateExpression(MockValidationException):
    invalid_update_expr_msg = "Invalid UpdateExpression: {update_expression_error}"

    def __init__(self, update_expression_error: str):
        self.update_expression_error = update_expression_error
        super().__init__(
            self.invalid_update_expr_msg.format(
                update_expression_error=update_expression_error
            )
        )


class InvalidConditionExpression(MockValidationException):
    invalid_condition_expr_msg = (
        "Invalid ConditionExpression: {condition_expression_error}"
    )

    def __init__(self, condition_expression_error: str):
        self.condition_expression_error = condition_expression_error
        super().__init__(
            self.invalid_condition_expr_msg.format(
                condition_expression_error=condition_expression_error
            )
        )


class ConditionAttributeIsReservedKeyword(InvalidConditionExpression):
    attribute_is_keyword_msg = (
        "Attribute name is a reserved keyword; reserved keyword: {keyword}"
    )

    def __init__(self, keyword: str):
        self.keyword = keyword
        super().__init__(self.attribute_is_keyword_msg.format(keyword=keyword))


class AttributeDoesNotExist(MockValidationException):
    attr_does_not_exist_msg = (
        "The provided expression refers to an attribute that does not exist in the item"
    )

    def __init__(self) -> None:
        super().__init__(self.attr_does_not_exist_msg)


class ProvidedKeyDoesNotExist(MockValidationException):
    provided_key_does_not_exist_msg = (
        "The provided key element does not match the schema"
    )

    def __init__(self) -> None:
        super().__init__(self.provided_key_does_not_exist_msg)


class ExpressionAttributeNameNotDefined(InvalidUpdateExpression):
    name_not_defined_msg = "An expression attribute name used in the document path is not defined; attribute name: {n}"

    def __init__(self, attribute_name: str):
        self.not_defined_attribute_name = attribute_name
        super().__init__(self.name_not_defined_msg.format(n=attribute_name))


class AttributeIsReservedKeyword(InvalidUpdateExpression):
    attribute_is_keyword_msg = (
        "Attribute name is a reserved keyword; reserved keyword: {keyword}"
    )

    def __init__(self, keyword: str):
        self.keyword = keyword
        super().__init__(self.attribute_is_keyword_msg.format(keyword=keyword))


class ExpressionAttributeValueNotDefined(InvalidUpdateExpression):
    attr_value_not_defined_msg = "An expression attribute value used in expression is not defined; attribute value: {attribute_value}"

    def __init__(self, attribute_value: str):
        self.attribute_value = attribute_value
        super().__init__(
            self.attr_value_not_defined_msg.format(attribute_value=attribute_value)
        )


class ExpressionAttributeValuesEmpty(MockValidationException):
    def __init__(self) -> None:
        super().__init__("ExpressionAttributeValues must not be empty")


class UpdateExprSyntaxError(InvalidUpdateExpression):
    update_expr_syntax_error_msg = "Syntax error; {error_detail}"

    def __init__(self, error_detail: str):
        self.error_detail = error_detail
        super().__init__(
            self.update_expr_syntax_error_msg.format(error_detail=error_detail)
        )


class InvalidTokenException(UpdateExprSyntaxError):
    token_detail_msg = 'token: "{token}", near: "{near}"'

    def __init__(self, token: str, near: str):
        self.token = token
        self.near = near
        super().__init__(self.token_detail_msg.format(token=token, near=near))


class InvalidExpressionAttributeNameKey(MockValidationException):
    invalid_expr_attr_name_msg = (
        'ExpressionAttributeNames contains invalid key: Syntax error; key: "{key}"'
    )

    def __init__(self, key: str):
        self.key = key
        super().__init__(self.invalid_expr_attr_name_msg.format(key=key))


class ItemSizeTooLarge(MockValidationException):
    item_size_too_large_msg = "Item size has exceeded the maximum allowed size"

    def __init__(self) -> None:
        super().__init__(self.item_size_too_large_msg)


class ItemSizeToUpdateTooLarge(MockValidationException):
    item_size_to_update_too_large_msg = (
        "Item size to update has exceeded the maximum allowed size"
    )

    def __init__(self) -> None:
        super().__init__(self.item_size_to_update_too_large_msg)


class HashKeyTooLong(MockValidationException):
    # deliberately no space between of and {lim}
    key_too_large_msg = f"One or more parameter values were invalid: Size of hashkey has exceeded the maximum size limit of{HASH_KEY_MAX_LENGTH} bytes"

    def __init__(self) -> None:
        super().__init__(self.key_too_large_msg)


class RangeKeyTooLong(MockValidationException):
    key_too_large_msg = f"One or more parameter values were invalid: Aggregated size of all range keys has exceeded the size limit of {RANGE_KEY_MAX_LENGTH} bytes"

    def __init__(self) -> None:
        super().__init__(self.key_too_large_msg)


class IncorrectOperandType(InvalidUpdateExpression):
    inv_operand_msg = "Incorrect operand type for operator or function; operator or function: {f}, operand type: {t}"

    def __init__(self, operator_or_function: str, operand_type: str):
        self.operator_or_function = operator_or_function
        self.operand_type = operand_type
        super().__init__(
            self.inv_operand_msg.format(f=operator_or_function, t=operand_type)
        )


class IncorrectDataType(MockValidationException):
    inc_data_type_msg = "An operand in the update expression has an incorrect data type"

    def __init__(self) -> None:
        super().__init__(self.inc_data_type_msg)


class ConditionalCheckFailed(DynamodbException):
    error_type = ERROR_TYPE_PREFIX + "ConditionalCheckFailedException"

    def __init__(
        self, msg: Optional[str] = None, item: Optional[Dict[str, Any]] = None
    ):
        _msg = msg or "The conditional request failed"
        super().__init__(ConditionalCheckFailed.error_type, _msg)
        if item:
            self.description = json.dumps(
                {
                    "__type": ConditionalCheckFailed.error_type,
                    # Note the uppercase Message
                    # This ensures the message is only part of the 'error': {'message': .., 'code': ..}
                    "Message": _msg,
                    "Item": item,
                }
            )
        else:
            self.description = json.dumps(
                {
                    "__type": ConditionalCheckFailed.error_type,
                    # Note that lowercase 'message'
                    # This ensures that 'message' is a top-level field in the response
                    # (in addition to being part of the 'error': {'message': .., 'code': ..}
                    "message": _msg,
                }
            )


class TransactionCanceledException(DynamodbException):
    cancel_reason_msg = "Transaction cancelled, please refer cancellation reasons for specific reasons [{}]"
    error_type = "com.amazonaws.dynamodb.v20120810#TransactionCanceledException"

    def __init__(self, errors: List[Any]):
        msg = self.cancel_reason_msg.format(
            ", ".join([str(code) for code, _, _ in errors])
        )
        super().__init__(
            error_type=TransactionCanceledException.error_type, message=msg
        )
        reasons = []
        for code, message, item in errors:
            r = {"Code": code, "Message": message} if code else {"Code": "None"}
            if item:
                r["Item"] = item
            reasons.append(r)

        self.description = json.dumps(
            {
                "__type": TransactionCanceledException.error_type,
                "CancellationReasons": reasons,
                "Message": msg,
            }
        )


class MultipleTransactionsException(MockValidationException):
    msg = "Transaction request cannot include multiple operations on one item"

    def __init__(self) -> None:
        super().__init__(self.msg)


class TooManyTransactionsException(MockValidationException):
    msg = (
        "1 validation error detected at 'transactItems' failed to satisfy constraint: "
        "Member must have length less than or equal to 100."
    )

    def __init__(self) -> None:
        super().__init__(self.msg)


class EmptyKeyAttributeException(MockValidationException):
    empty_str_msg = "One or more parameter values were invalid: An AttributeValue may not contain an empty string"
    # AWS has a different message for empty index keys
    empty_index_msg = "One or more parameter values are not valid. The update expression attempted to update a secondary index key to a value that is not supported. The AttributeValue for a key attribute cannot contain an empty string value."

    def __init__(self, key_in_index: bool = False):
        super().__init__(self.empty_index_msg if key_in_index else self.empty_str_msg)


class UpdateHashRangeKeyException(MockValidationException):
    msg = "One or more parameter values were invalid: Cannot update attribute {}. This attribute is part of the key"

    def __init__(self, key_name: str):
        super().__init__(self.msg.format(key_name))


class InvalidAttributeTypeError(MockValidationException):
    msg = "One or more parameter values were invalid: Type mismatch for key {} expected: {} actual: {}"

    def __init__(
        self, name: Optional[str], expected_type: Optional[str], actual_type: str
    ):
        super().__init__(self.msg.format(name, expected_type, actual_type))


class DuplicateUpdateExpression(InvalidUpdateExpression):
    def __init__(self, names: List[str]):
        super().__init__(
            f"Two document paths overlap with each other; must remove or rewrite one of these paths; path one: [{names[0]}], path two: [{names[1]}]"
        )


class InvalidProjectionExpression(MockValidationException):
    msg = (
        "Invalid ProjectionExpression: "
        "Two document paths overlap with each other; must remove or rewrite one of these paths; "
        "path one: [{paths[0]}], path two: [{paths[1]}]"
    )

    def __init__(self, paths: List[str]):
        super().__init__(self.msg.format(paths=paths))


class TooManyClauses(InvalidUpdateExpression):
    def __init__(self, _type: str) -> None:
        super().__init__(
            f'The "{_type}" section can only be used once in an update expression;'
        )


class ResourceNotFoundException(JsonRESTError):
    def __init__(self, msg: Optional[str] = None, table_name: Optional[str] = None):
        err = ERROR_TYPE_PREFIX + "ResourceNotFoundException"
        default_msg = "Requested resource not found"
        if table_name is not None:
            default_msg += f": Table: {table_name} not found"
        super().__init__(err, msg or default_msg)


class TableNotFoundException(JsonRESTError):
    def __init__(self, name: str):
        err = ERROR_TYPE_PREFIX + "TableNotFoundException"
        super().__init__(err, f"Table not found: {name}")


class PointInTimeRecoveryUnavailable(JsonRESTError):
    def __init__(self, name: str):
        err = ERROR_TYPE_PREFIX + "PointInTimeRecoveryUnavailableException"
        super().__init__(
            err, f"Point in time recovery is not enabled for table '{name}'"
        )


class SourceTableNotFoundException(JsonRESTError):
    def __init__(self, source_table_name: str):
        er = ERROR_TYPE_PREFIX + "SourceTableNotFoundException"
        super().__init__(er, f"Source table not found: {source_table_name}")


class BackupNotFoundException(JsonRESTError):
    def __init__(self, backup_arn: str):
        er = ERROR_TYPE_PREFIX + "BackupNotFoundException"
        super().__init__(er, f"Backup not found: {backup_arn}")


class TableAlreadyExistsException(JsonRESTError):
    def __init__(self, target_table_name: str):
        er = ERROR_TYPE_PREFIX + "TableAlreadyExistsException"
        super().__init__(er, f"Table already exists: {target_table_name}")


class ResourceInUseException(JsonRESTError):
    def __init__(self, msg: Optional[str] = None) -> None:
        er = ERROR_TYPE_PREFIX + "ResourceInUseException"
        super().__init__(er, msg or "Resource in use")


class StreamAlreadyEnabledException(JsonRESTError):
    def __init__(self) -> None:
        er = ERROR_TYPE_PREFIX + "ResourceInUseException"
        super().__init__(er, "Cannot enable stream")


class InvalidConversion(JsonRESTError):
    def __init__(self) -> None:
        er = "SerializationException"
        super().__init__(er, "NUMBER_VALUE cannot be converted to String")


class TransactWriteSingleOpException(MockValidationException):
    there_can_be_only_one = (
        "TransactItems can only contain one of Check, Put, Update or Delete"
    )

    def __init__(self) -> None:
        super().__init__(self.there_can_be_only_one)


class SerializationException(DynamodbException):
    def __init__(self, msg: str):
        super().__init__(error_type="SerializationException", message=msg)


class UnknownKeyType(MockValidationException):
    def __init__(self, key_type: str, position: str):
        msg = f"1 validation error detected: Value '{key_type}' at '{position}' failed to satisfy constraint: Member must satisfy enum value set: [HASH, RANGE]"
        super().__init__(msg)


class DeletionProtectedException(MockValidationException):
    def __init__(self, table_name: str):
        msg = f"1 validation error detected: Table '{table_name}' can't be deleted while DeletionProtectionEnabled is set to True"
        super().__init__(msg)


class PolicyNotFoundException(DynamodbException):
    error_type = ERROR_TYPE_PREFIX + "PolicyNotFoundException"

    def __init__(self, message: str):
        super().__init__(PolicyNotFoundException.error_type, message=message)
        self.exception_msg = message



from moto.core.utils import convert_regex_to_flask_path


def test_flask_path_converting_simple() -> None:
    assert convert_regex_to_flask_path("/") == "/"
    assert convert_regex_to_flask_path("/$") == "/"

    assert convert_regex_to_flask_path("/foo") == "/foo"

    assert convert_regex_to_flask_path("/foo/bar/") == "/foo/bar/"


def test_flask_path_converting_regex() -> None:
    assert (
        convert_regex_to_flask_path(r"/(?P<key_name>[a-zA-Z0-9\-_]+)")
        == r'/<regex("[a-zA-Z0-9\-_]+"):key_name>'
    )

    assert (
        convert_regex_to_flask_path(r"(?P<account_id>\d+)/(?P<queue_name>.*)$")
        == r'<regex("\d+"):account_id>/<regex(".*"):queue_name>'
    )



from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value import (
    JSONataTemplateValue,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class Output(EvalComponent):
    jsonata_template_value: JSONataTemplateValue

    def __init__(self, jsonata_template_value: JSONataTemplateValue):
        self.jsonata_template_value = jsonata_template_value

    def _eval_body(self, env: Environment) -> None:
        self.jsonata_template_value.eval(env=env)
        output_value = env.stack.pop()
        env.states.reset(input_value=output_value)



import logging
from typing import Any, Set

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceCondition,
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for

LOG = logging.getLogger(__name__)

_SUPPORTED_INTEGRATION_PATTERNS: Set[ResourceCondition] = {
    ResourceCondition.WaitForTaskToken,
}

# Defines bindings of lower-cased service names to the StepFunctions service name included in error messages.
_SERVICE_ERROR_NAMES = {"dynamodb": "DynamoDb", "sfn": "Sfn"}


class StateTaskServiceAwsSdk(StateTaskServiceCallback):
    def __init__(self):
        super().__init__(supported_integration_patterns=_SUPPORTED_INTEGRATION_PATTERNS)

    def from_state_props(self, state_props: StateProps) -> None:
        super().from_state_props(state_props=state_props)

    def _get_sfn_resource_type(self) -> str:
        return f"{self.resource.service_name}:{self.resource.api_name}"

    @staticmethod
    def _normalise_service_error_name(service_name: str) -> str:
        # Computes the normalised service error name for the given service.

        # Return the explicit binding if one exists.
        service_name_lower = service_name.lower()
        if service_name_lower in StateTaskServiceAwsSdk._SERVICE_ERROR_NAMES:
            return StateTaskServiceAwsSdk._SERVICE_ERROR_NAMES[service_name_lower]

        # Revert to returning the resource's service name and log the missing binding.
        LOG.error(
            f"No normalised service error name for aws-sdk integration was found for service: '{service_name}'"
        )
        return service_name

    @staticmethod
    def _normalise_exception_name(norm_service_name: str, ex: Exception) -> str:
        ex_name = ex.__class__.__name__
        norm_ex_name = f"{norm_service_name}.{norm_service_name if ex_name == 'ClientError' else ex_name}"
        if not norm_ex_name.endswith("Exception"):
            norm_ex_name += "Exception"
        return norm_ex_name

    def _get_task_failure_event(self, error: str, cause: str) -> FailureEvent:
        return FailureEvent(
            error_name=StatesErrorName(typ=StatesErrorNameType.StatesTaskFailed),
            event_type=HistoryEventType.TaskFailed,
            event_details=EventDetails(
                taskFailedEventDetails=TaskFailedEventDetails(
                    resource=self._get_sfn_resource(),
                    resourceType=self._get_sfn_resource_type(),
                    error=error,
                    cause=cause,
                )
            ),
        )

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            norm_service_name: str = self._normalise_service_error_name(
                self.resource.api_name
            )
            error: str = self._normalise_exception_name(norm_service_name, ex)

            error_message: str = ex.response["Error"]["Message"]
            cause_details = [
                f"Service: {norm_service_name}",
                f"Status Code: {ex.response['ResponseMetadata']['HTTPStatusCode']}",
                f"Request ID: {ex.response['ResponseMetadata']['RequestId']}",
            ]
            if "HostId" in ex.response["ResponseMetadata"]:
                cause_details.append(
                    f'Extended Request ID: {ex.response["ResponseMetadata"]["HostId"]}'
                )

            cause: str = f"{error_message} ({', '.join(cause_details)})"
            failure_event = self._get_task_failure_event(error=error, cause=cause)
            return failure_event
        return super()._from_error(env=env, ex=ex)

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: Any,
    ):
        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        api_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
        )
        response = getattr(api_client, api_action)(**normalised_parameters) or dict()

        env.stack.append(response)



# -*- coding: utf-8 -*-

# This data is a verbatim copy of that produced using the AWS CLI (requires AWS
# account) and jq utility to pull out the right part and reformat:
#
# aws polly describe-voices | jq .Voices --indent 4
VOICE_DATA = [
    {
        "Gender": "Female",
        "Id": "Isabelle",
        "LanguageCode": "fr-BE",
        "LanguageName": "Belgian French",
        "Name": "Isabelle",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Danielle",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Danielle",
        "SupportedEngines": ["long-form", "neural"],
    },
    {
        "Gender": "Male",
        "Id": "Gregory",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Gregory",
        "SupportedEngines": ["long-form", "neural"],
    },
    {
        "Gender": "Female",
        "Id": "Burcu",
        "LanguageCode": "tr-TR",
        "LanguageName": "Turkish",
        "Name": "Burcu",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Jitka",
        "LanguageCode": "cs-CZ",
        "LanguageName": "Czech",
        "Name": "Jitka",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Sabrina",
        "LanguageCode": "de-CH",
        "LanguageName": "Swiss Standard German",
        "Name": "Sabrina",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Kevin",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Kevin",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Filiz",
        "LanguageCode": "tr-TR",
        "LanguageName": "Turkish",
        "Name": "Filiz",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Elin",
        "LanguageCode": "sv-SE",
        "LanguageName": "Swedish",
        "Name": "Elin",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Astrid",
        "LanguageCode": "sv-SE",
        "LanguageName": "Swedish",
        "Name": "Astrid",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Tatyana",
        "LanguageCode": "ru-RU",
        "LanguageName": "Russian",
        "Name": "Tatyana",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Maxim",
        "LanguageCode": "ru-RU",
        "LanguageName": "Russian",
        "Name": "Maxim",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Carmen",
        "LanguageCode": "ro-RO",
        "LanguageName": "Romanian",
        "Name": "Carmen",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Ines",
        "LanguageCode": "pt-PT",
        "LanguageName": "Portuguese",
        "Name": "Inês",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Cristiano",
        "LanguageCode": "pt-PT",
        "LanguageName": "Portuguese",
        "Name": "Cristiano",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Vitoria",
        "LanguageCode": "pt-BR",
        "LanguageName": "Brazilian Portuguese",
        "Name": "Vitória",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Ricardo",
        "LanguageCode": "pt-BR",
        "LanguageName": "Brazilian Portuguese",
        "Name": "Ricardo",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Camila",
        "LanguageCode": "pt-BR",
        "LanguageName": "Brazilian Portuguese",
        "Name": "Camila",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Maja",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Name": "Maja",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Jan",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Name": "Jan",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Jacek",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Name": "Jacek",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Ewa",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Name": "Ewa",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Ola",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Name": "Ola",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Lisa",
        "LanguageCode": "nl-BE",
        "LanguageName": "Belgian Dutch",
        "Name": "Lisa",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Ruben",
        "LanguageCode": "nl-NL",
        "LanguageName": "Dutch",
        "Name": "Ruben",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Lotte",
        "LanguageCode": "nl-NL",
        "LanguageName": "Dutch",
        "Name": "Lotte",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Laura",
        "LanguageCode": "nl-NL",
        "LanguageName": "Dutch",
        "Name": "Laura",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Ida",
        "LanguageCode": "nb-NO",
        "LanguageName": "Norwegian",
        "Name": "Ida",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Liv",
        "LanguageCode": "nb-NO",
        "LanguageName": "Norwegian",
        "Name": "Liv",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Seoyeon",
        "LanguageCode": "ko-KR",
        "LanguageName": "Korean",
        "Name": "Seoyeon",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Kazuha",
        "LanguageCode": "ja-JP",
        "LanguageName": "Japanese",
        "Name": "Kazuha",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Tomoko",
        "LanguageCode": "ja-JP",
        "LanguageName": "Japanese",
        "Name": "Tomoko",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Takumi",
        "LanguageCode": "ja-JP",
        "LanguageName": "Japanese",
        "Name": "Takumi",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Mizuki",
        "LanguageCode": "ja-JP",
        "LanguageName": "Japanese",
        "Name": "Mizuki",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Bianca",
        "LanguageCode": "it-IT",
        "LanguageName": "Italian",
        "Name": "Bianca",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Giorgio",
        "LanguageCode": "it-IT",
        "LanguageName": "Italian",
        "Name": "Giorgio",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Carla",
        "LanguageCode": "it-IT",
        "LanguageName": "Italian",
        "Name": "Carla",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Karl",
        "LanguageCode": "is-IS",
        "LanguageName": "Icelandic",
        "Name": "Karl",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Dora",
        "LanguageCode": "is-IS",
        "LanguageName": "Icelandic",
        "Name": "Dóra",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Mathieu",
        "LanguageCode": "fr-FR",
        "LanguageName": "French",
        "Name": "Mathieu",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Lea",
        "LanguageCode": "fr-FR",
        "LanguageName": "French",
        "Name": "Léa",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Celine",
        "LanguageCode": "fr-FR",
        "LanguageName": "French",
        "Name": "Céline",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Chantal",
        "LanguageCode": "fr-CA",
        "LanguageName": "Canadian French",
        "Name": "Chantal",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Gabrielle",
        "LanguageCode": "fr-CA",
        "LanguageName": "Canadian French",
        "Name": "Gabrielle",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Penelope",
        "LanguageCode": "es-US",
        "LanguageName": "US Spanish",
        "Name": "Penélope",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Miguel",
        "LanguageCode": "es-US",
        "LanguageName": "US Spanish",
        "Name": "Miguel",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Lupe",
        "LanguageCode": "es-US",
        "LanguageName": "US Spanish",
        "Name": "Lupe",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Mia",
        "LanguageCode": "es-MX",
        "LanguageName": "Mexican Spanish",
        "Name": "Mia",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Lucia",
        "LanguageCode": "es-ES",
        "LanguageName": "Castilian Spanish",
        "Name": "Lucia",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Enrique",
        "LanguageCode": "es-ES",
        "LanguageName": "Castilian Spanish",
        "Name": "Enrique",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Conchita",
        "LanguageCode": "es-ES",
        "LanguageName": "Castilian Spanish",
        "Name": "Conchita",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Geraint",
        "LanguageCode": "en-GB-WLS",
        "LanguageName": "Welsh English",
        "Name": "Geraint",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Salli",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Salli",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Matthew",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Matthew",
        "SupportedEngines": ["generative", "neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Kimberly",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Kimberly",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Kendra",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Kendra",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Justin",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Justin",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Joey",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Joey",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Joanna",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Joanna",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Ivy",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Ivy",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Aria",
        "LanguageCode": "en-NZ",
        "LanguageName": "New Zealand English",
        "Name": "Aria",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Ayanda",
        "LanguageCode": "en-ZA",
        "LanguageName": "South African English",
        "Name": "Ayanda",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Raveena",
        "LanguageCode": "en-IN",
        "LanguageName": "Indian English",
        "Name": "Raveena",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Aditi",
        "LanguageCode": "en-IN",
        "LanguageName": "Indian English",
        "Name": "Aditi",
        "AdditionalLanguageCodes": ["hi-IN"],
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Emma",
        "LanguageCode": "en-GB",
        "LanguageName": "British English",
        "Name": "Emma",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Brian",
        "LanguageCode": "en-GB",
        "LanguageName": "British English",
        "Name": "Brian",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Amy",
        "LanguageCode": "en-GB",
        "LanguageName": "British English",
        "Name": "Amy",
        "SupportedEngines": ["generative", "neural", "standard"],
    },
    {
        "Gender": "Male",
        "Id": "Russell",
        "LanguageCode": "en-AU",
        "LanguageName": "Australian English",
        "Name": "Russell",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Nicole",
        "LanguageCode": "en-AU",
        "LanguageName": "Australian English",
        "Name": "Nicole",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Olivia",
        "LanguageCode": "en-AU",
        "LanguageName": "Australian English",
        "Name": "Olivia",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Vicki",
        "LanguageCode": "de-DE",
        "LanguageName": "German",
        "Name": "Vicki",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Marlene",
        "LanguageCode": "de-DE",
        "LanguageName": "German",
        "Name": "Marlene",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Hans",
        "LanguageCode": "de-DE",
        "LanguageName": "German",
        "Name": "Hans",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Naja",
        "LanguageCode": "da-DK",
        "LanguageName": "Danish",
        "Name": "Naja",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Male",
        "Id": "Mads",
        "LanguageCode": "da-DK",
        "LanguageName": "Danish",
        "Name": "Mads",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Sofie",
        "LanguageCode": "da-DK",
        "LanguageName": "Danish",
        "Name": "Sofie",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Gwyneth",
        "LanguageCode": "cy-GB",
        "LanguageName": "Welsh",
        "Name": "Gwyneth",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Zhiyu",
        "LanguageCode": "cmn-CN",
        "LanguageName": "Chinese Mandarin",
        "Name": "Zhiyu",
        "SupportedEngines": ["neural", "standard"],
    },
    {
        "Gender": "Female",
        "Id": "Zeina",
        "LanguageCode": "arb",
        "LanguageName": "Arabic",
        "Name": "Zeina",
        "SupportedEngines": ["standard"],
    },
    {
        "Gender": "Female",
        "Id": "Hala",
        "LanguageCode": "ar-AE",
        "LanguageName": "Gulf Arabic",
        "Name": "Hala",
        "AdditionalLanguageCodes": ["arb"],
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Arlet",
        "LanguageCode": "ca-ES",
        "LanguageName": "Catalan",
        "Name": "Arlet",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Hannah",
        "LanguageCode": "de-AT",
        "LanguageName": "Austrian German",
        "Name": "Hannah",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Ruth",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Ruth",
        "SupportedEngines": ["generative", "long-form", "neural"],
    },
    {
        "Gender": "Male",
        "Id": "Stephen",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Name": "Stephen",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Kajal",
        "LanguageCode": "en-IN",
        "LanguageName": "Indian English",
        "Name": "Kajal",
        "AdditionalLanguageCodes": ["hi-IN"],
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Hiujin",
        "LanguageCode": "yue-CN",
        "LanguageName": "Cantonese",
        "Name": "Hiujin",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Suvi",
        "LanguageCode": "fi-FI",
        "LanguageName": "Finnish",
        "Name": "Suvi",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Female",
        "Id": "Niamh",
        "LanguageCode": "en-IE",
        "LanguageName": "Irish English",
        "Name": "Niamh",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Arthur",
        "LanguageCode": "en-GB",
        "LanguageName": "British English",
        "Name": "Arthur",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Daniel",
        "LanguageCode": "de-DE",
        "LanguageName": "German",
        "Name": "Daniel",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Liam",
        "LanguageCode": "fr-CA",
        "LanguageName": "Canadian French",
        "Name": "Liam",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Pedro",
        "LanguageCode": "es-US",
        "LanguageName": "US Spanish",
        "Name": "Pedro",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Sergio",
        "LanguageCode": "es-ES",
        "LanguageName": "Castilian Spanish",
        "Name": "Sergio",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Andres",
        "LanguageCode": "es-MX",
        "LanguageName": "Mexican Spanish",
        "Name": "Andrés",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Remi",
        "LanguageCode": "fr-FR",
        "LanguageName": "French",
        "Name": "Rémi",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Adriano",
        "LanguageCode": "it-IT",
        "LanguageName": "Italian",
        "Name": "Adriano",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Thiago",
        "LanguageCode": "pt-BR",
        "LanguageName": "Brazilian Portuguese",
        "Name": "Thiago",
        "SupportedEngines": ["neural"],
    },
    {
        "Gender": "Male",
        "Id": "Zayd",
        "LanguageCode": "ar-AE",
        "LanguageName": "Gulf Arabic",
        "Name": "Zayd",
        "AdditionalLanguageCodes": ["arb"],
        "SupportedEngines": ["neural"],
    },
]

LANGUAGE_CODES = {voice["LanguageCode"] for voice in VOICE_DATA}

VOICE_IDS = {voice["Name"] for voice in VOICE_DATA}



import boto3
import pytest

from moto import mock_aws

from . import application_autoscaling_aws_verified

target_tracking_config = {
    "TargetTrackingScalingPolicyConfiguration": {
        "TargetValue": 70.0,
        "PredefinedMetricSpecification": {
            "PredefinedMetricType": "DynamoDBReadCapacityUtilization"
        },
    }
}
step_scaling_config = {
    "StepScalingPolicyConfiguration": {
        "AdjustmentType": "ChangeInCapacity",
        "StepAdjustments": [{"ScalingAdjustment": 10}],
        "MinAdjustmentMagnitude": 2,
    }
}


@pytest.mark.parametrize(
    ["policy_type", "policy_body_kwargs"],
    [
        ["TargetTrackingScaling", target_tracking_config],
        ["TargetTrackingScaling", step_scaling_config],
    ],
)
@application_autoscaling_aws_verified()
def test_put_scaling_policy_with_unknown_policytype(
    policy_type,
    policy_body_kwargs,
    client=None,
    namespace=None,
    resource_id=None,
    scalable_dimension=None,
):
    with pytest.raises(client.exceptions.ValidationException) as exc:
        client.put_scaling_policy(
            PolicyName="MyPolicy",
            ServiceNamespace=namespace,
            ResourceId=resource_id,
            ScalableDimension=scalable_dimension,
            PolicyType="ABCDEFG",
            **policy_body_kwargs,
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "1 validation error detected: Value 'ABCDEFG' at 'policyType' failed to satisfy constraint: Member must satisfy enum value set: [PredictiveScaling, StepScaling, TargetTrackingScaling]"
    )


@pytest.mark.parametrize(
    ["policy_type", "policy_body_kwargs"],
    [["TargetTrackingScaling", target_tracking_config]],
)
@application_autoscaling_aws_verified()
def test_describe_scaling_policy_alarms(
    policy_type,
    policy_body_kwargs,
    client=None,
    namespace=None,
    resource_id=None,
    scalable_dimension=None,
):
    policy_name = "MyPolicy"

    response = client.put_scaling_policy(
        PolicyName=policy_name,
        ServiceNamespace=namespace,
        ResourceId=resource_id,
        ScalableDimension=scalable_dimension,
        PolicyType=policy_type,
        **policy_body_kwargs,
    )
    assert len(response["Alarms"]) == 4

    alarms = client.describe_scaling_policies(
        PolicyNames=[policy_name],
        ServiceNamespace=namespace,
        ResourceId=resource_id,
        ScalableDimension=scalable_dimension,
    )["ScalingPolicies"][0]["Alarms"]
    assert len(alarms) == 4

    cloudwatch = boto3.client("cloudwatch", "us-east-1")
    for alarm in alarms:
        cw_alarm = cloudwatch.describe_alarms(AlarmNames=[alarm["AlarmName"]])[
            "MetricAlarms"
        ][0]
        assert cw_alarm["AlarmName"] == alarm["AlarmName"]
        assert cw_alarm["AlarmArn"] == alarm["AlarmARN"]

    client.delete_scaling_policy(
        PolicyName=policy_name,
        ServiceNamespace=namespace,
        ResourceId=resource_id,
        ScalableDimension=scalable_dimension,
    )

    for alarm in response["Alarms"]:
        assert (
            cloudwatch.describe_alarms(AlarmNames=[alarm["AlarmName"]])["MetricAlarms"]
            == []
        )


@mock_aws
def test_describe_scaling_policy_with_ecs_alarms():
    policy_name = "MyPolicy"

    create_ecs_service("test", "servicename")

    namespace = "ecs"
    resource_id = "service/test/servicename"
    scalable_dimension = "ecs:service:DesiredCount"

    client = boto3.client("application-autoscaling", region_name="us-east-1")

    client.register_scalable_target(
        ServiceNamespace=namespace,
        ResourceId=resource_id,
        ScalableDimension=scalable_dimension,
        MinCapacity=1,
        MaxCapacity=4,
    )

    client.put_scaling_policy(
        PolicyName=policy_name,
        PolicyType="TargetTrackingScaling",
        ResourceId=resource_id,
        ScalableDimension=scalable_dimension,
        ServiceNamespace=namespace,
        TargetTrackingScalingPolicyConfiguration={
            "PredefinedMetricSpecification": {
                "PredefinedMetricType": "ECSServiceAverageMemoryUtilization",
            },
            "ScaleInCooldown": 2,
            "ScaleOutCooldown": 12,
            "TargetValue": 6,
        },
    )

    alarms = client.describe_scaling_policies(
        ServiceNamespace="ecs", ResourceId=resource_id
    )["ScalingPolicies"][0]["Alarms"]

    assert len(alarms) == 2

    cloudwatch = boto3.client("cloudwatch", "us-east-1")
    for alarm in alarms:
        cw_alarm = cloudwatch.describe_alarms(AlarmNames=[alarm["AlarmName"]])[
            "MetricAlarms"
        ][0]
        assert f"TargetTracking-{resource_id}" in cw_alarm["AlarmName"]
        assert {"Name": "ClusterName", "Value": "test"} in cw_alarm["Dimensions"]


def create_ecs_service(cluster_name, service_name):
    client = boto3.client("ecs", region_name="us-east-1")
    client.create_cluster(clusterName=cluster_name)

    client.register_task_definition(
        family="test_ecs_task",
        containerDefinitions=[
            {
                "name": "hello_world",
                "image": "docker/hello-world:latest",
                "cpu": 1024,
                "memory": 400,
                "essential": True,
                "environment": [
                    {"name": "AWS_ACCESS_KEY_ID", "value": "SOME_ACCESS_KEY"}
                ],
                "logConfiguration": {"logDriver": "json-file"},
            }
        ],
    )
    client.create_service(
        cluster=cluster_name,
        serviceName=service_name,
        taskDefinition="test_ecs_task",
        desiredCount=2,
        platformVersion="2",
    )



"""Unit tests for backup-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_create_backup_plan():
    client = boto3.client("backup", region_name="eu-west-1")
    response = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
    )
    resp = client.create_backup_plan(
        BackupPlan={
            "BackupPlanName": "backupplan-foobar",
            "Rules": [
                {
                    "RuleName": "foobar",
                    "TargetBackupVaultName": response["BackupVaultName"],
                },
            ],
        },
    )
    assert "BackupPlanId" in resp
    assert "BackupPlanArn" in resp
    assert "CreationDate" in resp
    assert "VersionId" in resp


@mock_aws
def test_create_backup_plan_already_exists():
    client = boto3.client("backup", region_name="eu-west-1")
    backup_plan_name = "backup_plan_foobar"
    rules = [
        {
            "RuleName": "foobar",
            "TargetBackupVaultName": "backup-vault-foobar",
        },
    ]
    client.create_backup_plan(
        BackupPlan={"BackupPlanName": backup_plan_name, "Rules": rules}
    )

    with pytest.raises(ClientError) as exc:
        client.create_backup_plan(
            BackupPlan={"BackupPlanName": backup_plan_name, "Rules": rules}
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "AlreadyExistsException"


@mock_aws
def test_get_backup_plan():
    client = boto3.client("backup", region_name="eu-west-1")
    response = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
    )
    plan = client.create_backup_plan(
        BackupPlan={
            "BackupPlanName": "backupplan-foobar",
            "Rules": [
                {
                    "RuleName": "foobar",
                    "TargetBackupVaultName": response["BackupVaultName"],
                },
            ],
        },
    )
    resp = client.get_backup_plan(
        BackupPlanId=plan["BackupPlanId"], VersionId=plan["VersionId"]
    )
    assert "BackupPlan" in resp


@mock_aws
def test_get_backup_plan_invalid_id():
    client = boto3.client("backup", region_name="eu-west-1")

    with pytest.raises(ClientError) as exc:
        client.get_backup_plan(BackupPlanId="foobar")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_get_backup_plan_invalid_version_id():
    client = boto3.client("backup", region_name="eu-west-1")

    plan = client.create_backup_plan(
        BackupPlan={
            "BackupPlanName": "backupplan-foobar",
            "Rules": [
                {
                    "RuleName": "foobar",
                    "TargetBackupVaultName": "Backup-vault-foobar",
                },
            ],
        },
    )
    with pytest.raises(ClientError) as exc:
        client.get_backup_plan(BackupPlanId=plan["BackupPlanId"], VersionId="foobar")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_get_backup_plan_with_multiple_rules():
    client = boto3.client("backup", region_name="eu-west-1")
    plan = client.create_backup_plan(
        BackupPlan={
            "BackupPlanName": "backupplan-foobar",
            "Rules": [
                {
                    "RuleName": "rule1",
                    "TargetBackupVaultName": "backupvault-foobar",
                    "ScheduleExpression": "cron(0 1 ? * * *)",
                    "StartWindowMinutes": 60,
                    "CompletionWindowMinutes": 120,
                },
                {
                    "RuleName": "rule2",
                    "TargetBackupVaultName": "backupvault-foobar",
                },
            ],
        },
    )
    resp = client.get_backup_plan(BackupPlanId=plan["BackupPlanId"])
    for rule in resp["BackupPlan"]["Rules"]:
        assert "ScheduleExpression" in rule
        assert "StartWindowMinutes" in rule
        assert "CompletionWindowMinutes" in rule
        assert "RuleId" in rule


@mock_aws
def test_delete_backup_plan():
    client = boto3.client("backup", region_name="eu-west-1")
    response = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
    )
    plan = client.create_backup_plan(
        BackupPlan={
            "BackupPlanName": "backupplan-foobar",
            "Rules": [
                {
                    "RuleName": "foobar",
                    "TargetBackupVaultName": response["BackupVaultName"],
                },
            ],
        },
    )

    resp = client.delete_backup_plan(BackupPlanId=plan["BackupPlanId"])
    assert "BackupPlanId" in resp
    assert "BackupPlanArn" in resp
    assert "DeletionDate" in resp
    assert "VersionId" in resp

    resp = client.get_backup_plan(
        BackupPlanId=plan["BackupPlanId"], VersionId=plan["VersionId"]
    )
    assert "DeletionDate" in resp


@mock_aws
def test_delete_backup_plan_invalid_id():
    client = boto3.client("backup", region_name="eu-west-1")

    with pytest.raises(ClientError) as exc:
        client.delete_backup_plan(BackupPlanId="foobar")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_backup_plans():
    client = boto3.client("backup", region_name="eu-west-1")
    for i in range(1, 3):
        client.create_backup_plan(
            BackupPlan={
                "BackupPlanName": f"backup-plan-{i}",
                "Rules": [
                    {
                        "RuleName": "foobar",
                        "TargetBackupVaultName": "backupvault-foobar",
                    },
                ],
            },
        )
    resp = client.list_backup_plans()
    backup_plans = resp["BackupPlansList"]
    assert backup_plans[0]["BackupPlanName"] == "backup-plan-1"
    assert backup_plans[1]["BackupPlanName"] == "backup-plan-2"
    assert len(backup_plans) == 2
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200


@mock_aws
def test_list_backup_plans_without_include_deleted():
    client = boto3.client("backup", region_name="eu-west-1")

    for i in range(1, 3):
        client.create_backup_plan(
            BackupPlan={
                "BackupPlanName": f"backup-plan-{i}",
                "Rules": [
                    {
                        "RuleName": "foobar",
                        "TargetBackupVaultName": "backupvault-foobar",
                    },
                ],
            },
        )
    resp = client.list_backup_plans()
    client.delete_backup_plan(BackupPlanId=resp["BackupPlansList"][0]["BackupPlanId"])
    resp_list = client.list_backup_plans()
    backup_plans = resp_list["BackupPlansList"]
    assert backup_plans[0]["BackupPlanName"] == "backup-plan-2"
    assert len(backup_plans) == 1


@mock_aws
def test_list_backup_plans_with_include_deleted():
    client = boto3.client("backup", region_name="eu-west-1")
    for i in range(1, 3):
        client.create_backup_plan(
            BackupPlan={
                "BackupPlanName": f"backup-plan-{i}",
                "Rules": [
                    {
                        "RuleName": "foobar",
                        "TargetBackupVaultName": "backupvault-foobar",
                    },
                ],
            },
        )
    resp = client.list_backup_plans()
    client.delete_backup_plan(BackupPlanId=resp["BackupPlansList"][0]["BackupPlanId"])
    resp_list = client.list_backup_plans(IncludeDeleted=True)
    backup_plans = resp_list["BackupPlansList"]
    assert backup_plans[0]["BackupPlanName"] == "backup-plan-1"
    assert backup_plans[1]["BackupPlanName"] == "backup-plan-2"
    assert len(backup_plans) == 2


@mock_aws
def test_create_backup_vault():
    client = boto3.client("backup", region_name="eu-west-1")
    resp = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
        BackupVaultTags={
            "foo": "bar",
        },
    )
    assert "BackupVaultName" in resp
    assert "BackupVaultArn" in resp
    assert "CreationDate" in resp


@mock_aws
def test_create_backup_vault_already_exists():
    client = boto3.client("backup", region_name="eu-west-1")
    backup_vault_name = "backup_vault_foobar"
    client.create_backup_vault(BackupVaultName=backup_vault_name)

    with pytest.raises(ClientError) as exc:
        client.create_backup_vault(BackupVaultName=backup_vault_name)
    err = exc.value.response["Error"]
    assert err["Code"] == "AlreadyExistsException"


@mock_aws
def test_list_backup_vaults():
    client = boto3.client("backup", region_name="eu-west-1")
    for i in range(1, 3):
        client.create_backup_vault(
            BackupVaultName=f"backup-vault-{i}",
        )
    resp = client.list_backup_vaults()
    backup_plans = resp["BackupVaultList"]
    assert backup_plans[0]["BackupVaultName"] == "backup-vault-1"
    assert backup_plans[1]["BackupVaultName"] == "backup-vault-2"
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200


@mock_aws
def test_list_tags_vault():
    client = boto3.client("backup", region_name="eu-west-1")
    vault = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
        BackupVaultTags={
            "key1": "value1",
            "key2": "value2",
        },
    )
    resp = client.list_tags(ResourceArn=vault["BackupVaultArn"])
    assert resp["Tags"] == {"key1": "value1", "key2": "value2"}


@mock_aws
def test_list_tags_plan():
    client = boto3.client("backup", region_name="eu-west-1")
    response = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
    )
    plan = client.create_backup_plan(
        BackupPlan={
            "BackupPlanName": "backupplan-foobar",
            "Rules": [
                {
                    "RuleName": "foobar",
                    "TargetBackupVaultName": response["BackupVaultName"],
                },
            ],
        },
        BackupPlanTags={
            "key1": "value1",
            "key2": "value2",
        },
    )
    resp = client.list_tags(ResourceArn=plan["BackupPlanArn"])
    assert resp["Tags"] == {"key1": "value1", "key2": "value2"}


@mock_aws
def test_tag_resource():
    client = boto3.client("backup", region_name="eu-west-1")
    vault = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
        BackupVaultTags={
            "key1": "value1",
        },
    )
    resource_arn = vault["BackupVaultArn"]
    client.tag_resource(
        ResourceArn=resource_arn, Tags={"key2": "value2", "key3": "value3"}
    )
    resp = client.list_tags(ResourceArn=resource_arn)
    assert resp["Tags"] == {"key1": "value1", "key2": "value2", "key3": "value3"}


@mock_aws
def test_untag_resource():
    client = boto3.client("backup", region_name="eu-west-1")
    vault = client.create_backup_vault(
        BackupVaultName="backupvault-foobar",
        BackupVaultTags={
            "key1": "value1",
        },
    )
    resource_arn = vault["BackupVaultArn"]
    client.tag_resource(
        ResourceArn=resource_arn, Tags={"key2": "value2", "key3": "value3"}
    )
    resp = client.untag_resource(ResourceArn=resource_arn, TagKeyList=["key2"])
    resp = client.list_tags(ResourceArn=resource_arn)
    assert resp["Tags"] == {"key1": "value1", "key3": "value3"}



from __future__ import annotations

import datetime
import json
import logging
from typing import Dict, Optional

from moto.stepfunctions.parser.api import (
    Arn,
    CloudWatchEventsExecutionDataDetails,
    DescribeExecutionOutput,
    DescribeStateMachineForExecutionOutput,
    ExecutionListItem,
    ExecutionStatus,
    GetExecutionHistoryOutput,
    HistoryEventList,
    InvalidName,
    SensitiveCause,
    SensitiveError,
    StartExecutionOutput,
    StartSyncExecutionOutput,
    StateMachineType,
    SyncExecutionStatus,
    Timestamp,
    TraceHeader,
    VariableReferences,
)
from moto.stepfunctions.parser.asl.eval.evaluation_details import (
    AWSExecutionDetails,
    EvaluationDetails,
    ExecutionDetails,
    StateMachineDetails,
)
from moto.stepfunctions.parser.asl.eval.event.logging import (
    CloudWatchLoggingSession,
)
from moto.stepfunctions.parser.asl.eval.program_state import (
    ProgramEnded,
    ProgramError,
    ProgramState,
    ProgramStopped,
    ProgramTimedOut,
)
from moto.stepfunctions.parser.asl.static_analyser.variable_references_static_analyser import (
    VariableReferencesStaticAnalyser,
)
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.backend.activity import Activity
from moto.stepfunctions.parser.backend.execution_worker import (
    ExecutionWorker,
    SyncExecutionWorker,
)
from moto.stepfunctions.parser.backend.execution_worker_comm import (
    ExecutionWorkerCommunication,
)
from moto.stepfunctions.parser.backend.state_machine import (
    StateMachineInstance,
    StateMachineVersion,
)

LOG = logging.getLogger(__name__)


class BaseExecutionWorkerCommunication(ExecutionWorkerCommunication):
    execution: Execution

    def __init__(self, execution: Execution):
        self.execution = execution

    def _reflect_execution_status(self):
        exit_program_state: ProgramState = (
            self.execution.exec_worker.env.program_state()
        )
        self.execution.stop_date = datetime.datetime.now(tz=datetime.timezone.utc)
        if isinstance(exit_program_state, ProgramEnded):
            self.execution.exec_status = ExecutionStatus.SUCCEEDED
            self.execution.output = self.execution.exec_worker.env.states.get_input()
        elif isinstance(exit_program_state, ProgramStopped):
            self.execution.exec_status = ExecutionStatus.ABORTED
        elif isinstance(exit_program_state, ProgramError):
            self.execution.exec_status = ExecutionStatus.FAILED
            self.execution.error = exit_program_state.error.get("error")
            self.execution.cause = exit_program_state.error.get("cause")
        elif isinstance(exit_program_state, ProgramTimedOut):
            self.execution.exec_status = ExecutionStatus.TIMED_OUT
        else:
            raise RuntimeWarning(
                f"Execution ended with unsupported ProgramState type '{type(exit_program_state)}'."
            )

    def terminated(self) -> None:
        self._reflect_execution_status()


class Execution:
    name: str
    sm_type: StateMachineType
    role_arn: Arn
    exec_arn: Arn

    account_id: str
    region_name: str

    state_machine: StateMachineInstance
    start_date: Timestamp
    input_data: Optional[json]
    input_details: Optional[CloudWatchEventsExecutionDataDetails]
    trace_header: Optional[TraceHeader]
    _cloud_watch_logging_session: Optional[CloudWatchLoggingSession]

    exec_status: Optional[ExecutionStatus]
    stop_date: Optional[Timestamp]

    output: Optional[json]
    output_details: Optional[CloudWatchEventsExecutionDataDetails]

    error: Optional[SensitiveError]
    cause: Optional[SensitiveCause]

    exec_worker: Optional[ExecutionWorker]

    _activity_store: Dict[Arn, Activity]

    def __init__(
        self,
        name: str,
        sm_type: StateMachineType,
        role_arn: Arn,
        exec_arn: Arn,
        account_id: str,
        region_name: str,
        state_machine: StateMachineInstance,
        start_date: Timestamp,
        cloud_watch_logging_session: Optional[CloudWatchLoggingSession],
        activity_store: Dict[Arn, Activity],
        input_data: Optional[json] = None,
        trace_header: Optional[TraceHeader] = None,
    ):
        self.name = name
        self.sm_type = sm_type
        self.role_arn = role_arn
        self.exec_arn = exec_arn
        self.execution_arn = exec_arn
        self.account_id = account_id
        self.region_name = region_name
        self.state_machine = state_machine
        self._cloud_watch_logging_session = cloud_watch_logging_session
        self.input_data = input_data
        self.input_details = CloudWatchEventsExecutionDataDetails(included=True)
        self.trace_header = trace_header
        self.exec_status = None
        self.stop_date = None
        self.output = None
        self.output_details = CloudWatchEventsExecutionDataDetails(included=True)
        self.exec_worker = None
        self.error = None
        self.cause = None
        self._activity_store = activity_store

        # Compatibility with mock SFN
        self.state_machine_arn = state_machine.arn
        self.start_date = start_date
        self.execution_input = input_data

    @property
    def status(self):
        return self.exec_status.value

    def to_start_output(self) -> StartExecutionOutput:
        return StartExecutionOutput(
            executionArn=self.exec_arn, startDate=self.start_date
        )

    def to_describe_output(self) -> DescribeExecutionOutput:
        describe_output = DescribeExecutionOutput(
            executionArn=self.exec_arn,
            stateMachineArn=self.state_machine.arn,
            name=self.name,
            status=self.exec_status,
            startDate=self.start_date,
            stopDate=self.stop_date,
            input=to_json_str(self.input_data, separators=(",", ":")),
            inputDetails=self.input_details,
            traceHeader=self.trace_header,
        )
        if describe_output["status"] == ExecutionStatus.SUCCEEDED:
            describe_output["output"] = to_json_str(self.output, separators=(",", ":"))
            describe_output["outputDetails"] = self.output_details
        if self.error is not None:
            describe_output["error"] = self.error
        if self.cause is not None:
            describe_output["cause"] = self.cause
        return describe_output

    def to_describe_state_machine_for_execution_output(
        self,
    ) -> DescribeStateMachineForExecutionOutput:
        state_machine: StateMachineInstance = self.state_machine
        state_machine_arn = (
            state_machine.source_arn
            if isinstance(state_machine, StateMachineVersion)
            else state_machine.arn
        )
        out = DescribeStateMachineForExecutionOutput(
            stateMachineArn=state_machine_arn,
            name=state_machine.name,
            definition=state_machine.definition,
            roleArn=self.role_arn,
            # The date and time the state machine associated with an execution was updated.
            updateDate=state_machine.create_date,
            loggingConfiguration=state_machine.logging_config,
        )
        revision_id = self.state_machine.revision_id
        if self.state_machine.revision_id:
            out["revisionId"] = revision_id
        variable_references: VariableReferences = (
            VariableReferencesStaticAnalyser.process_and_get(
                definition=self.state_machine.definition
            )
        )
        if variable_references:
            out["variableReferences"] = variable_references
        return out

    def to_execution_list_item(self) -> ExecutionListItem:
        if isinstance(self.state_machine, StateMachineVersion):
            state_machine_arn = self.state_machine.source_arn
            state_machine_version_arn = self.state_machine.arn
        else:
            state_machine_arn = self.state_machine.arn
            state_machine_version_arn = None

        item = ExecutionListItem(
            executionArn=self.exec_arn,
            stateMachineArn=state_machine_arn,
            name=self.name,
            status=self.exec_status,
            startDate=self.start_date,
            stopDate=self.stop_date,
        )
        if state_machine_version_arn is not None:
            item["stateMachineVersionArn"] = state_machine_version_arn
        return item

    def to_history_output(self) -> GetExecutionHistoryOutput:
        env = self.exec_worker.env
        event_history: HistoryEventList = list()
        if env is not None:
            # The execution has not started yet.
            event_history: HistoryEventList = env.event_manager.get_event_history()
        return GetExecutionHistoryOutput(events=event_history)

    def _get_start_execution_worker_comm(self) -> BaseExecutionWorkerCommunication:
        return BaseExecutionWorkerCommunication(self)

    def _get_start_aws_execution_details(self) -> AWSExecutionDetails:
        return AWSExecutionDetails(
            account=self.account_id, region=self.region_name, role_arn=self.role_arn
        )

    def get_start_execution_details(self) -> ExecutionDetails:
        return ExecutionDetails(
            arn=self.exec_arn,
            name=self.name,
            role_arn=self.role_arn,
            inpt=self.input_data,
            start_time=self.start_date,
        )

    def get_start_state_machine_details(self) -> StateMachineDetails:
        return StateMachineDetails(
            arn=self.state_machine.arn,
            name=self.state_machine.name,
            typ=self.state_machine.sm_type,
            definition=self.state_machine.definition,
        )

    def _get_start_execution_worker(self) -> ExecutionWorker:
        return ExecutionWorker(
            evaluation_details=EvaluationDetails(
                aws_execution_details=self._get_start_aws_execution_details(),
                execution_details=self.get_start_execution_details(),
                state_machine_details=self.get_start_state_machine_details(),
            ),
            exec_comm=self._get_start_execution_worker_comm(),
            cloud_watch_logging_session=self._cloud_watch_logging_session,
            activity_store=self._activity_store,
        )

    def start(self) -> None:
        # TODO: checks exec_worker does not exists already?
        if self.exec_worker:
            raise InvalidName()  # TODO.
        self.exec_worker = self._get_start_execution_worker()
        self.exec_status = ExecutionStatus.RUNNING
        self.exec_worker.start()

    def stop(
        self, stop_date: datetime.datetime, error: Optional[str], cause: Optional[str]
    ):
        exec_worker: Optional[ExecutionWorker] = self.exec_worker
        if exec_worker:
            exec_worker.stop(stop_date=stop_date, cause=cause, error=error)


class SyncExecutionWorkerCommunication(BaseExecutionWorkerCommunication):
    execution: SyncExecution

    def _reflect_execution_status(self) -> None:
        super()._reflect_execution_status()
        exit_status: ExecutionStatus = self.execution.exec_status
        if exit_status == ExecutionStatus.SUCCEEDED:
            self.execution.sync_execution_status = SyncExecutionStatus.SUCCEEDED
        elif exit_status == ExecutionStatus.TIMED_OUT:
            self.execution.sync_execution_status = SyncExecutionStatus.TIMED_OUT
        else:
            self.execution.sync_execution_status = SyncExecutionStatus.FAILED


class SyncExecution(Execution):
    sync_execution_status: Optional[SyncExecutionStatus] = None

    def _get_start_execution_worker(self) -> SyncExecutionWorker:
        return SyncExecutionWorker(
            evaluation_details=EvaluationDetails(
                aws_execution_details=self._get_start_aws_execution_details(),
                execution_details=self.get_start_execution_details(),
                state_machine_details=self.get_start_state_machine_details(),
            ),
            exec_comm=self._get_start_execution_worker_comm(),
            cloud_watch_logging_session=self._cloud_watch_logging_session,
            activity_store=self._activity_store,
        )

    def _get_start_execution_worker_comm(self) -> BaseExecutionWorkerCommunication:
        return SyncExecutionWorkerCommunication(self)

    def to_start_sync_execution_output(self) -> StartSyncExecutionOutput:
        start_output = StartSyncExecutionOutput(
            executionArn=self.exec_arn,
            stateMachineArn=self.state_machine.arn,
            name=self.name,
            status=self.sync_execution_status,
            startDate=self.start_date,
            stopDate=self.stop_date,
            input=to_json_str(self.input_data, separators=(",", ":")),
            inputDetails=self.input_details,
            traceHeader=self.trace_header,
        )
        if self.sync_execution_status == SyncExecutionStatus.SUCCEEDED:
            start_output["output"] = to_json_str(self.output, separators=(",", ":"))
        if self.output_details:
            start_output["outputDetails"] = self.output_details
        if self.error is not None:
            start_output["error"] = self.error
        if self.cause is not None:
            start_output["cause"] = self.cause
        return start_output



import unittest

import boto3

from moto import mock_aws


@mock_aws
class TestLogsTagging(unittest.TestCase):
    def setUp(self) -> None:
        self.logs = boto3.client("logs", region_name="us-east-2")
        self.rtapi = boto3.client("resourcegroupstaggingapi", region_name="us-east-2")
        self.resources_tagged = []
        self.resources_untagged = []
        for i in range(3):
            self.logs.create_log_group(logGroupName=f"test{i}", tags={"key1": "val1"})
        self.arns = [
            lg["logGroupArn"] for lg in self.logs.describe_log_groups()["logGroups"]
        ]

    def test_get_resources_logs(self):
        resp = self.rtapi.get_resources(ResourceTypeFilters=["logs"])
        assert len(resp["ResourceTagMappingList"]) == 3

        resp = self.rtapi.get_resources(ResourceTypeFilters=["logs:loggroup"])
        assert len(resp["ResourceTagMappingList"]) == 3

    def test_tag_resources_logs(self):
        # WHEN
        # we tag resources
        self.rtapi.tag_resources(
            ResourceARNList=self.arns,
            Tags={"key2": "val2"},
        )

        # THEN
        # we can retrieve the tags using the RDS API
        def get_tags(arn):
            return self.logs.list_tags_for_resource(resourceArn=arn)["tags"]

        for arn in self.arns:
            assert get_tags(arn) == {"key1": "val1", "key2": "val2"}
